from kratos import *


class SW_NET(Generator):
    '''
    The switching network will bring in multiple items per cycle and direct
    them to the appropriate aggregation buffer
    '''
    def __init__(self,
                 interconnect_in,
                 offsets,  # Offset between the interconnect_in ports
                 data_width,
                 memory_width,  # Will be the size of the agg buffer
                 num_aggregators,
                 num_banks,
                 stride_0):
        super().__init__("sw_net")

        self.interconnect_in = interconnect_in
        self.offsets = offsets
        self.data_width = data_width
        self.memory_width = memory_width
        self.num_aggregators = num_aggregators
        self.num_banks = num_banks
        self.stride_0 = stride_0

        # PORT DEFS: begin
        self._clk = self.clock("clk")
        self._clk_en = self.input("clk_en", 1)
        self._reset = self.reset("reset")
        self._flush = self.input("flush", 1)

        self._data_in = self.input("data_in",
                                   data_width,
                                   size=interconnect_in,
                                   packed=True,
                                   explicit_array=True)  # Actually an array
        self._valid_in = self.input("valid_in", interconnect_in)

        self._data_out  # should be interconnect_in as well - basically a single stage input pipe
        self._agg_index_out  # should be the num_aggregators
        self._valid_out  # Should be the num of banks

        # Somehow get data to the output ->
        # probably with an address -> which will be 0-num_aggregators
        # Then tag it with valid or not



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import *
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.utils.util import safe_wire, add_counter, decode
import kratos as kts


class StrgUBSRAMOnly(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 tb_height=2):

        super().__init__("strg_ub_sram_only")

        ##################################################################################
        # Capture constructor parameter...
        ##################################################################################
        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.mem_width = mem_width
        self.tb_height = tb_height
        self.mem_depth = mem_depth
        self.config_width = config_width
        self.data_width = data_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support

        self.default_iterator_support = 6
        self.default_config_width = 16
        self.sram_iterator_support = 6
        self.agg_rd_addr_gen_width = 8

        ##################################################################################
        # IO
        ##################################################################################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._cycle_count = self.input("cycle_count", 16)

        # agg to sram for loop
        self._floop_mux_sel = self.input("floop_mux_sel",
                                         width=max(clog2(self.default_iterator_support), 1),
                                         size=self.interconnect_input_ports,
                                         explicit_array=True,
                                         packed=True)

        self._floop_restart = self.input("floop_restart",
                                         width=1,
                                         size=self.interconnect_input_ports,
                                         explicit_array=True,
                                         packed=True)

        # sram to tb for loop
        self._loops_sram2tb_mux_sel = self.input("loops_sram2tb_mux_sel",
                                                 width=max(clog2(self.default_iterator_support), 1),
                                                 size=self.interconnect_output_ports,
                                                 explicit_array=True,
                                                 packed=True)

        self._loops_sram2tb_restart = self.input("loops_sram2tb_restart",
                                                 width=1,
                                                 size=self.interconnect_output_ports,
                                                 explicit_array=True,
                                                 packed=True)

        self._agg_read = self.input("agg_read", self.interconnect_input_ports)
        self._t_read = self.input("t_read", self.interconnect_output_ports)

        # data from aggs, get decoded for sram_write_data which is wired to data_to_sram
        self._agg_data_out = self.input(f"agg_data_out", self.data_width,
                                        size=(self.interconnect_input_ports,
                                              self.fetch_width),
                                        packed=True,
                                        explicit_array=True)
        self._agg_data_out.add_attribute(FormalAttr(self._agg_data_out.name, FormalSignalConstraint.SEQUENCE, "agg"))
        # sram attribute for data_in, comes from cut gen of agg_only for agg_data_out_top

        self._wen_to_sram = self.output("wen_to_sram", 1, packed=True)
        self._cen_to_sram = self.output("cen_to_sram", 1, packed=True)
        self._addr_to_sram = self.output("addr_to_sram", clog2(self.mem_depth), packed=True)
        self._data_to_sram = self.output("data_to_sram", self.data_width,
                                         size=self.fetch_width,
                                         packed=True)

        ##################################################################################
        # INTERNAL SIGNALS
        ##################################################################################
        self._s_write_addr = self.var("s_write_addr", self.config_width,
                                      size=self.interconnect_input_ports,
                                      packed=True,
                                      explicit_array=True)

        self._s_read_addr = self.var("s_read_addr",
                                     self.config_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        self._write = self.var("write", 1)
        self._read = self.var("read", 1)
        self._addr = self.var("addr", clog2(self.mem_depth))

        self._sram_write_data = self.var("sram_write_data", data_width,
                                         size=self.fetch_width,
                                         packed=True)

        self.mem_addr_width = clog2(self.mem_depth)

        for i in range(self.interconnect_input_ports):

            _AG = AddrGen(iterator_support=self.default_iterator_support,
                          config_width=self.mem_addr_width)
            self.add_child(f"input_addr_gen_{i}",
                           _AG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_read[i],
                           # mux_sel=self._floop_mux_sel[i],
                           restart=self._floop_restart[i])
            safe_wire(gen=self, w_to=_AG.ports.mux_sel, w_from=self._floop_mux_sel[i])
            safe_wire(gen=self, w_to=self._s_write_addr[i], w_from=_AG.ports.addr_out)

        ##################################################################################
        # TB PATHS
        ##################################################################################
        for i in range(self.interconnect_output_ports):

            _AG = AddrGen(iterator_support=self.default_iterator_support,
                          config_width=self.mem_addr_width)
            self.add_child(f"output_addr_gen_{i}",
                           _AG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._t_read[i],
                           # mux_sel=self._loops_sram2tb_mux_sel[i],
                           restart=self._loops_sram2tb_restart[i])
            safe_wire(gen=self, w_to=_AG.ports.mux_sel, w_from=self._loops_sram2tb_mux_sel[i])
            safe_wire(gen=self, w_to=self._s_read_addr[i], w_from=_AG.ports.addr_out)

        ##################################################################################
        # WIRE TO SRAM INTERFACE
        ##################################################################################
        # Now select the write address as a decode of the underlying enables
        self.wire(self._addr_to_sram, self._addr)
        self.wire(self._data_to_sram, self._sram_write_data)
        self.wire(self._wen_to_sram, self._write)
        self.wire(self._cen_to_sram, self._write | self._read)

        self.wire(self._write, self._agg_read.r_or())
        self.wire(self._read, self._t_read.r_or())

        self.wire(self._sram_write_data, decode(self, self._agg_read, self._agg_data_out))

        self._write_addr = decode(self, self._agg_read, self._s_write_addr)
        self._read_addr = decode(self, self._t_read, self._s_read_addr)
        self.add_code(self.set_sram_addr)

    @always_comb
    def set_sram_addr(self):
        if self._write:
            self._addr = self._write_addr[clog2(self.mem_depth) - 1, 0]
        else:
            self._addr = self._read_addr[clog2(self.mem_depth) - 1, 0]


if __name__ == "__main__":
    lake_dut = StrgUBSRAMOnly()
    verilog(lake_dut, filename="strg_ub_sram_only.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import *
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.modules.agg_only import StrgUBAggOnly
from lake.modules.agg_sram_shared import StrgUBAggSRAMShared
from lake.modules.sram_only import StrgUBSRAMOnly
from lake.modules.sram_tb_shared import StrgUBSRAMTBShared
from lake.modules.tb_only import StrgUBTBOnly
from lake.utils.util import safe_wire, add_counter, decode
import kratos as kts


class StrgUBVec(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 tb_height=2,
                 agg_data_top=False):

        super().__init__("strg_ub_vec")

        ##################################################################################
        # Capture constructor parameter...
        ##################################################################################
        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.tb_height = tb_height
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.banks = banks
        self.config_width = config_width
        self.data_width = data_width
        self.read_delay = read_delay
        self.rw_same_cycle = rw_same_cycle
        self.input_config_width = config_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support

        self.input_iterator_support = 6
        self.output_iterator_support = 6
        self.default_iterator_support = 6
        self.default_config_width = 16
        self.sram_iterator_support = 6
        self.agg_rd_addr_gen_width = 8

        ##################################################################################
        # IO
        ##################################################################################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._data_in = self.input("data_in", self.data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        self._data_from_sram = self.input("data_from_strg", self.data_width,
                                          size=self.fetch_width,
                                          packed=True)

        self._wen_to_sram = self.output("wen_to_strg", 1, packed=True)
        self._cen_to_sram = self.output("cen_to_strg", 1, packed=True)
        self._addr_to_sram = self.output("addr_out", clog2(self.mem_depth), packed=True)
        self._data_to_sram = self.output("data_to_strg", self.data_width,
                                         size=self.fetch_width,
                                         packed=True)

        self._valid_out = self.output("accessor_output", self.interconnect_output_ports)
        self._data_out = self.output("data_out", self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        ##################################################################################
        # CYCLE COUNTER
        ##################################################################################

        # Create cycle counter to share...
        self._cycle_count = add_counter(self, "cycle_count", 16)

        agg_only = StrgUBAggOnly(data_width=self.data_width,
                                 mem_width=self.mem_width,
                                 mem_depth=self.mem_depth,
                                 banks=self.banks,
                                 input_addr_iterator_support=self.input_iterator_support,
                                 output_addr_iterator_support=self.output_iterator_support,
                                 input_sched_iterator_support=self.input_iterator_support,
                                 output_sched_iterator_support=self.output_iterator_support,
                                 interconnect_input_ports=self.interconnect_input_ports,
                                 interconnect_output_ports=self.interconnect_output_ports,
                                 read_delay=self.read_delay,
                                 rw_same_cycle=self.rw_same_cycle,
                                 agg_height=self.agg_height,
                                 config_width=self.input_config_width)

        agg_sram_shared = StrgUBAggSRAMShared(data_width=self.data_width,
                                              mem_width=self.mem_width,
                                              mem_depth=self.mem_depth,
                                              banks=self.banks,
                                              input_addr_iterator_support=self.input_iterator_support,
                                              output_addr_iterator_support=self.output_iterator_support,
                                              input_sched_iterator_support=self.input_iterator_support,
                                              output_sched_iterator_support=self.output_iterator_support,
                                              interconnect_input_ports=self.interconnect_input_ports,
                                              interconnect_output_ports=self.interconnect_output_ports,
                                              read_delay=self.read_delay,
                                              rw_same_cycle=self.rw_same_cycle,
                                              agg_height=self.agg_height,
                                              config_width=self.input_config_width)

        sram_only = StrgUBSRAMOnly(data_width=self.data_width,
                                   mem_width=self.mem_width,
                                   mem_depth=self.mem_depth,
                                   banks=self.banks,
                                   input_addr_iterator_support=self.input_iterator_support,
                                   output_addr_iterator_support=self.output_iterator_support,
                                   input_sched_iterator_support=self.input_iterator_support,
                                   output_sched_iterator_support=self.output_iterator_support,
                                   interconnect_input_ports=self.interconnect_input_ports,
                                   interconnect_output_ports=self.interconnect_output_ports,
                                   read_delay=self.read_delay,
                                   rw_same_cycle=self.rw_same_cycle,
                                   agg_height=self.agg_height,
                                   config_width=self.input_config_width)

        sram_tb_shared = StrgUBSRAMTBShared(data_width=self.data_width,
                                            mem_width=self.mem_width,
                                            mem_depth=self.mem_depth,
                                            banks=self.banks,
                                            input_addr_iterator_support=self.input_iterator_support,
                                            output_addr_iterator_support=self.output_iterator_support,
                                            input_sched_iterator_support=self.input_iterator_support,
                                            output_sched_iterator_support=self.output_iterator_support,
                                            interconnect_input_ports=self.interconnect_input_ports,
                                            interconnect_output_ports=self.interconnect_output_ports,
                                            read_delay=self.read_delay,
                                            rw_same_cycle=self.rw_same_cycle,
                                            agg_height=self.agg_height,
                                            config_width=self.input_config_width)

        tb_only = StrgUBTBOnly(data_width=self.data_width,
                               mem_width=self.mem_width,
                               mem_depth=self.mem_depth,
                               banks=self.banks,
                               input_addr_iterator_support=self.input_iterator_support,
                               output_addr_iterator_support=self.output_iterator_support,
                               input_sched_iterator_support=self.input_iterator_support,
                               output_sched_iterator_support=self.output_iterator_support,
                               interconnect_input_ports=self.interconnect_input_ports,
                               interconnect_output_ports=self.interconnect_output_ports,
                               read_delay=self.read_delay,
                               rw_same_cycle=self.rw_same_cycle,
                               agg_height=self.agg_height,
                               config_width=self.input_config_width)

        self.add_child("agg_only",
                       agg_only,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       data_in=self._data_in)

        self.add_child("agg_sram_shared",
                       agg_sram_shared,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count)

        self.add_child("sram_only",
                       sram_only,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       wen_to_sram=self._wen_to_sram,
                       cen_to_sram=self._cen_to_sram,
                       addr_to_sram=self._addr_to_sram,
                       data_to_sram=self._data_to_sram)

        self.add_child("sram_tb_shared",
                       sram_tb_shared,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count)

        self.add_child("tb_only",
                       tb_only,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       sram_read_data=self._data_from_sram,
                       accessor_output=self._valid_out,
                       data_out=self._data_out)

        self.wire(agg_only.ports.agg_read, agg_sram_shared.ports.agg_read_out)
        self.wire(agg_only.ports.floop_mux_sel, agg_sram_shared.ports.floop_mux_sel)
        self.wire(agg_only.ports.floop_restart, agg_sram_shared.ports.floop_restart)

        self.wire(sram_only.ports.floop_mux_sel, agg_sram_shared.ports.floop_mux_sel)
        self.wire(sram_only.ports.floop_restart, agg_sram_shared.ports.floop_restart)
        self.wire(sram_only.ports.loops_sram2tb_mux_sel, sram_tb_shared.ports.loops_sram2tb_mux_sel)
        self.wire(sram_only.ports.loops_sram2tb_restart, sram_tb_shared.ports.loops_sram2tb_restart)
        self.wire(sram_only.ports.agg_read, agg_sram_shared.ports.agg_read_out)
        self.wire(sram_only.ports.t_read, sram_tb_shared.ports.t_read_out)
        self.wire(sram_only.ports.agg_data_out, agg_only.ports.agg_data_out)

        self.wire(tb_only.ports.t_read, sram_tb_shared.ports.t_read_out)
        self.wire(tb_only.ports.loops_sram2tb_mux_sel, sram_tb_shared.ports.loops_sram2tb_mux_sel)
        self.wire(tb_only.ports.loops_sram2tb_restart, sram_tb_shared.ports.loops_sram2tb_restart)

        if agg_data_top:
            self._agg_data_out = self.output(f"strg_ub_agg_data_out", self.data_width,
                                             size=(self.interconnect_input_ports,
                                                   self.fetch_width),
                                             packed=True,
                                             explicit_array=True)
            self.wire(self._agg_data_out, agg_only.ports.agg_data_out)


if __name__ == "__main__":
    lake_dut = StrgUBVec()
    verilog(lake_dut, filename="strg_ub_vec.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from math import log
from lake.attributes.sram_port_attr import SRAMPortAttr


class SRAMStubGenerator(Generator):

    ##########################
    # Generation             #
    ##########################
    def __init__(self,
                 sram_name,
                 data_width,
                 width_mult,
                 depth):
        super().__init__(sram_name)

        # for provided external sram macro
        self.external = True

        self.data_width = data_width
        self.width_mult = width_mult
        self.depth = depth

        ############################
        # Clock and Reset          #
        ############################
        self._clk = self.clock("sram_clk")
        self._clk.add_attribute(SRAMPortAttr("sram clk"))

        ############################
        # Inputs                   #
        ############################

        # attribute indicates that all these ports will be renamed to match
        # the port names for external sram macro
        self._wen = self.input("sram_wen", 1)
        self._wen.add_attribute(SRAMPortAttr("sram wen"))

        self._cen = self.input("sram_cen", 1)
        self._cen.add_attribute(SRAMPortAttr("sram cen"))

        self._addr = self.input("sram_addr", clog2(self.depth))
        self._addr.add_attribute(SRAMPortAttr("sram addr"))

        self._data_in = self.input("sram_data_in",
                                   self.data_width * self.width_mult)
        self._data_in.add_attribute(SRAMPortAttr("sram data in"))

        ############################
        # Outputs                  #
        ############################
        self._data_out = self.output("sram_data_out",
                                     self.data_width * self.width_mult)
        self._data_out.add_attribute(SRAMPortAttr("sram data out"))

        self._wtsel = self.input("sram_wtsel", 2)
        self._wtsel.add_attribute(SRAMPortAttr("sram wtsel"))

        self._rtsel = self.input("sram_rtsel", 2)
        self._rtsel.add_attribute(SRAMPortAttr("sram rtsel"))


if __name__ == "__main__":
    # to see interface, mark self.use_stub = True and self.external = False
    dut = SRAMStubGenerator(sram_name="TSMC",
                            data_width=16,
                            width_mult=1,
                            depth=124)
    verilog(dut, filename="tsmc_macro.sv")



from lake.models.model import Model
from lake.models.addr_gen_model import AddrGenModel
import math as mt
import kratos as kts
import numpy as np


class InputAddrCtrlModel(Model):

    def __init__(self,
                 interconnect_input_ports,
                 mem_depth,
                 banks,
                 num_tiles,
                 iterator_support,
                 max_port_schedule,
                 address_width,
                 data_width,
                 fetch_width):

        self.interconnect_input_ports = interconnect_input_ports
        self.mem_depth = mem_depth
        self.banks = banks
        self.num_tiles = num_tiles
        self.iterator_support = iterator_support
        self.address_width = address_width
        self.max_port_schedule = max_port_schedule
        self.data_width = data_width
        self.fetch_width = fetch_width
        self.fw_int = int(self.fetch_width / self.data_width)

        self.config = {}

        # Create child address generators
        self.addr_gens = []
        for i in range(self.interconnect_input_ports):
            new_addr_gen = AddrGenModel(iterator_support=self.iterator_support,
                                        address_width=self.address_width)
            self.addr_gens.append(new_addr_gen)

        self.mem_addr_width = kts.clog2(self.num_tiles * self.mem_depth)

        # Get local list of addresses
        self.addresses = []
        self.data_out = []
        for i in range(self.banks):
            self.addresses.append(0)
            self.data_out.append(0)

        # Initialize the configuration
        for i in range(self.interconnect_input_ports):
            self.config[f"address_gen_{i}_starting_addr"] = 0
            self.config[f"address_gen_{i}_dimensionality"] = 0
            for j in range(self.iterator_support):
                self.config[f"address_gen_{i}_strides_{j}"] = 0
                self.config[f"address_gen_{i}_ranges_{j}"] = 0

        # Set up the wen
        self.wen = []
        self.mem_addresses = []
        # self.port_sels = []
        for i in range(self.banks):
            self.wen.append(0)
            self.mem_addresses.append(0)
            # self.port_sels.append(0)

        self.port_out_exp = []
        for i in range(self.banks):
            inter = []
            for j in range(self.interconnect_input_ports):
                inter.append(0)
            self.port_out_exp.append(inter)

    def set_config(self, new_config):
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val
        # Configure children
        for i in range(self.interconnect_input_ports):
            addr_gen_config = {}
            addr_gen_config["starting_addr"] = self.config[f"address_gen_{i}_starting_addr"]
            addr_gen_config["dimensionality"] = self.config[f"address_gen_{i}_dimensionality"]
            for j in range(self.iterator_support):
                addr_gen_config[f"strides_{j}"] = self.config[f"address_gen_{i}_strides_{j}"]
                addr_gen_config[f"ranges_{j}"] = self.config[f"address_gen_{i}_ranges_{j}"]
            self.addr_gens[i].set_config(addr_gen_config)

    # Retrieve the current addresses from each generator
    def interact(self, valid, data_in, wen_en):
        '''
        Returns (wen, data, addrs, port_out)
        '''
        wen = self.get_wen(valid, wen_en)
        data_out = self.get_data_out(valid, data_in)
        port_out = self.get_port_out(valid)
        addrs = self.get_addrs(valid)
        self.step_addrs(valid)
        return(wen, data_out, addrs, port_out)

    def get_port_out(self, valid):
        for i in range(self.banks):
            for j in range(self.interconnect_input_ports):
                if (valid[j]):
                    self.port_out_exp[i][j] = 1
                else:
                    self.port_out_exp[i][j] = 0

        port_out = []
        for i in range(self.interconnect_input_ports):
            new_tmp = []
            for j in range(self.banks):
                new_tmp.append(self.port_out_exp[j][i])

            concat = np.bitwise_or.reduce(new_tmp)
            port_out.append(concat)

        return port_out

    def get_addrs(self, valid):
        for i in range(self.banks):
            self.addresses[i] = 0
        for i in range(self.interconnect_input_ports):
            if (valid[i]):
                to_get = self.addr_gens[i]
                if(self.banks == 1):
                    self.addresses[0] = to_get.get_address()
                    break
                else:
                    self.addresses[to_get.get_address() >> (self.mem_addr_width)] = to_get.get_address()
        return self.addresses

    def get_data_out(self, valid, data_in):
        assert len(data_in) == self.interconnect_input_ports, "Should feed proper length data"
        for i in range(self.banks):
            self.data_out[i] = [0 for z in range(self.fw_int)]
            # self.data_out[i] = [0] * len(data_in[0])
        for i in range(self.interconnect_input_ports):
            if (valid[i]):
                to_get = self.addr_gens[i]
                if(self.banks == 1):
                    self.data_out[0] = data_in[i]
                    break
                else:
                    self.data_out[to_get.get_address() >> (self.mem_addr_width)] = data_in[i]
        return self.data_out

    # Get the wen for the current valid input
    def get_wen(self, valid, wen_en):
        for i in range(self.banks):
            self.wen[i] = 0
        for i in range(self.interconnect_input_ports):
            if (valid[i]):
                to_get = self.addr_gens[i]
                if(self.banks == 1):
                    self.wen[0] = wen_en[i]
                    break
                else:
                    self.wen[to_get.get_address() >> (self.mem_addr_width)] = wen_en[i]
        return self.wen

    # Step the addresses based on valid
    def step_addrs(self, valid):
        for i, valid_input in enumerate(valid):
            if valid_input:
                to_step = self.addr_gens[i]
                to_step.step()

        # Not implemented
    def update_ports(self):
        raise NotImplementedError

    def peek(self):
        raise NotImplementedError



from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr


class ConfigReg(Generator):
    def __init__(self,
                 name,
                 data_width,
                 init_val):
        super().__init__(f"config_reg_{name}_{data_width}")

        self.add_attribute(ConfigRegAttr())

        self.init_val = init_val
        self.data_width = data_width

        self._out = self.output("o_data_out", data_width)
        self._clk = self.clock("i_clk")
        self._rst_n = self.reset("i_rst_n")

        self.add_code(self.set_out)

    @always((posedge, "i_clk"), (negedge, "i_rst_n"))
    def set_out(self):
        if ~self._rst_n:
            self._out = self.init_val


if __name__ == "__main__":
    data_width = 16
    configregdut = ConfigReg(name="config_test", data_width=data_width, init_val=2)
    verilog(configregdut, filename=f"config_reg_{data_width}.sv")



from kratos import *
import kratos as kts


def create_port_pkt(data_width,
                    consumer_ports):
    return PackedStruct(f"port_pkt_{data_width}_{consumer_ports}",
                        [("data", data_width, False),
                         ("port", consumer_ports, False),
                         ("valid", 1, False)])


class RWArbiter(Generator):
    '''
    The read_write arbiter dynamically schedules reads and writes to
    a single bank of SRAM
    '''
    def __init__(self,
                 fetch_width=16,
                 data_width=16,
                 memory_depth=32,
                 num_tiles=1,
                 int_in_ports=2,
                 int_out_ports=2,
                 strg_wr_ports=2,
                 strg_rd_ports=2,
                 read_delay=0,
                 rw_same_cycle=True,
                 separate_addresses=True):

        assert not (memory_depth & (memory_depth - 1)), "Memory depth needs to be a power of 2"

        super().__init__("rw_arbiter")
        # Absorb inputs
        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.int_in_ports = int_in_ports
        self.int_out_ports = int_out_ports
        self.strg_wr_ports = strg_wr_ports
        self.strg_rd_ports = strg_rd_ports
        self.memory_depth = memory_depth
        self.num_tiles = num_tiles
        self.mem_addr_width = clog2(self.num_tiles * self.memory_depth)
        self.read_delay = read_delay
        self.rw_same_cycle = rw_same_cycle
        self.separate_addresses = separate_addresses

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Generate the packed struct of the right size
        port_pkt_struct = create_port_pkt(self.fetch_width, self.int_out_ports)

        # Inputs
        self._wen_in = self.input("wen_in", self.strg_wr_ports)
        # self._wen_en = self.input("wen_en", self.strg_wr_ports)
        self._w_data = self.input("w_data",
                                  self.data_width,
                                  size=(self.strg_wr_ports,
                                        self.fw_int),
                                  explicit_array=True,
                                  packed=True)
        self._w_addr = self.input("w_addr", self.mem_addr_width,
                                  size=self.strg_wr_ports,
                                  explicit_array=True,
                                  packed=True)

        self._data_from_mem = self.input("data_from_mem",
                                         self.data_width,
                                         size=(self.strg_rd_ports,
                                               self.fw_int),
                                         explicit_array=True,
                                         packed=True)

        self._mem_valid_data = self.input("mem_valid_data",
                                          self.strg_rd_ports)
        self._out_mem_valid_data = self.output("out_mem_valid_data",
                                               self.strg_rd_ports)

        self._ren_in = self.input("ren_in", self.int_out_ports)
        self._ren_en = self.input("ren_en", self.int_out_ports)
        self._rd_addr = self.input("rd_addr",
                                   self.mem_addr_width,
                                   size=self.int_out_ports,
                                   explicit_array=True,
                                   packed=True)

        self._rd_addr_sel = self.var("rd_addr_sel",
                                     self.mem_addr_width,
                                     size=self.strg_rd_ports,
                                     explicit_array=True,
                                     packed=True)

        # Outputs
        self._out_data = self.output("out_data",
                                     self.data_width,
                                     size=(self.strg_rd_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)
        self._out_port = self.output("out_port", self.int_out_ports,
                                     size=self.strg_rd_ports,
                                     explicit_array=True,
                                     packed=True)
        self._out_valid = self.output("out_valid", self.strg_rd_ports)

        self._cen_mem = self.output("cen_mem", self.strg_rd_ports)
        self._wen_mem = self.output("wen_mem", self.strg_wr_ports)
        self._data_to_mem = self.output("data_to_mem",
                                        self.data_width,
                                        size=(self.strg_wr_ports,
                                              self.fw_int),
                                        explicit_array=True,
                                        packed=True)
        # In this case, need separate addresses
        if self.separate_addresses:
            self._wr_addr_to_mem = self.output("wr_addr_to_mem",
                                               self.mem_addr_width,
                                               size=self.strg_wr_ports,
                                               explicit_array=True,
                                               packed=True)
            self._rd_addr_to_mem = self.output("rd_addr_to_mem",
                                               self.mem_addr_width,
                                               size=self.strg_rd_ports,
                                               explicit_array=True,
                                               packed=True)
        # If the addresses are combined, we better have in==out
        else:
            assert self.strg_rd_ports == self.strg_wr_ports, \
                "Cannot have coalesced address with mismatch port count"
            assert not self.rw_same_cycle, \
                "Cannot read and write with a shared address...set rw_same_cycle to false"
            self._addr_to_mem = self.output("addr_to_mem", self.mem_addr_width,
                                            size=self.strg_rd_ports,
                                            explicit_array=True,
                                            packed=True)

        self._out_ack = self.output("out_ack", self.int_out_ports)

        # Local
        # self._rd_data = self.var("rd_data", self.fetch_width)
        self._wen_int = self.var("wen_int", self.strg_wr_ports)
        self._ren_int = self.var("ren_int", self.int_out_ports)
        self.wire(self._ren_int, self._ren_in & self._ren_en)
        self.wire(self._wen_int, self._wen_in)  # & self._wen_en)

        self._rd_valid = self.var("rd_valid", self.strg_rd_ports)
        self._rd_port = self.var("rd_port", self.int_out_ports,
                                 size=self.strg_rd_ports,
                                 explicit_array=True,
                                 packed=True)
        self._next_rd_port = self.var("next_rd_port", self.int_out_ports,
                                      size=self.strg_rd_ports,
                                      explicit_array=True,
                                      packed=True)

        # For demuxing the read ports
        self._done = self.var("done", self.strg_rd_ports)
        self.add_code(self.set_next_read_port_lowest)
        if(self.strg_rd_ports > 1):
            self._idx_cnt = self.var("idx_cnt", 5,
                                     size=self.strg_rd_ports - 1,
                                     explicit_array=True,
                                     packed=True)
            for i in range(self.strg_rd_ports - 1):
                self.add_code(self.set_next_read_port_alt, index=i + 1)

        # If we have more than one read port, we need to use slightly different logic
        # to set the other reads...

        self._next_rd_port_red = self.var("next_rd_port_red", self.int_out_ports)
        for i in range(self.int_out_ports):
            temp_port = self._next_rd_port[0][i]
            for j in range(self.strg_rd_ports - 1):
                temp_port = kts.concat(temp_port, self._next_rd_port[j + 1][i])
            self.wire(self._next_rd_port_red[i], temp_port.r_or())

        # The next read port can be used to acknowledge reads
        # We do not need to gate the ack if we can read and write in the same cycle
        if self.rw_same_cycle:
            self.wire(self._out_ack,
                      self._next_rd_port_red)
        else:
            self.wire(self._out_ack,
                      self._next_rd_port_red & kts.concat(*([~self._wen_int] * self._out_ack.width)))

        # self.add_code(self.mem_controls)
        if self.separate_addresses:
            for i in range(self.strg_wr_ports):
                self.add_code(self.mem_controls_wr, idx=i)
            for i in range(self.strg_rd_ports):
                self.add_code(self.mem_controls_rd, idx=i)
        else:
            for i in range(self.strg_rd_ports):
                self.add_code(self.mem_controls_combined, idx=i)

        if self.read_delay == 1:
            for i in range(self.strg_rd_ports):
                self.add_code(self.next_read_valid, idx=i)
        else:
            for i in range(self.strg_rd_ports):
                self.add_code(self.zero_delay_read, idx=i)
        self.add_code(self.output_stage)

    @always_comb
    def mem_controls_wr(self, idx):
        self._wen_mem[idx] = self._wen_int[idx]
        self._data_to_mem[idx] = self._w_data[idx]
        self._wr_addr_to_mem[idx] = self._w_addr[idx]

    @always_comb
    def mem_controls_rd(self, idx):
        # cen_mem acts as ren_mem when the ports are separated
        self._cen_mem[idx] = self._next_rd_port[idx].r_or()
        self._rd_addr_to_mem[idx] = self._rd_addr_sel[idx]

    @always_comb
    # Prioritizes writes over reads
    def mem_controls_combined(self, idx):
        self._wen_mem[idx] = self._wen_int[idx]
        self._cen_mem[idx] = (self._wen_int[idx] | (self._next_rd_port[idx].r_or()))
        self._data_to_mem[idx] = self._w_data[idx]
        # Consume wr over read
        if(self._wen_int[idx]):
            self._addr_to_mem[idx] = self._w_addr[idx]
        else:
            self._addr_to_mem[idx] = self._rd_addr_sel[idx]

    @always_comb
    # Find lowest ready
    def set_next_read_port_lowest(self):
        self._next_rd_port[0] = 0
        self._rd_addr_sel[0] = 0
        self._done[0] = 0
        for i in range(self.int_out_ports):
            if ~self._done[0]:
                if self._ren_int[i]:
                    self._rd_addr_sel[0] = self._rd_addr[i]
                    self._next_rd_port[0][i] = 1
                    self._done[0] = 1

    # Find lowest ready
    @always_comb
    def set_next_read_port_alt(self, index):
        self._next_rd_port[index] = 0
        self._idx_cnt[index - 1] = 0
        self._rd_addr_sel[index] = 0
        self._done[index] = 0
        for i in range(self.int_out_ports):
            if ~self._done[index]:
                if self._ren_int[i] & (self._idx_cnt[index - 1] == index):
                    self._done[index] = 1
                    self._rd_addr_sel[index] = self._rd_addr[i]
                    self._next_rd_port[index][i] = 1
                self._idx_cnt[index - 1] = self._idx_cnt[index - 1] + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def next_read_valid(self, idx):
        if ~self._rst_n:
            self._rd_port[idx] = 0
            self._rd_valid[idx] = 0
        else:
            self._rd_valid[idx] = ((~self._wen_int[idx] | (self.rw_same_cycle)) &
                                   (self._next_rd_port[idx].r_or()))
            self._rd_port[idx] = self._next_rd_port[idx]

    @always_comb
    def zero_delay_read(self, idx):
        self._rd_valid[idx] = (~self._wen_int[idx] | (self.rw_same_cycle)) & (self._next_rd_port[idx].r_or())
        self._rd_port[idx] = self._next_rd_port[idx]

    @always_comb
    def output_stage(self):
        self._out_data = self._data_from_mem
        self._out_port = self._rd_port
        self._out_valid = self._rd_valid
        self._out_mem_valid_data = self._mem_valid_data


if __name__ == "__main__":
    db_dut = RWArbiter()

    verilog(db_dut, filename="rw_arbiter.sv")



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.utils.util import safe_wire, add_counter, decode
import kratos as kts


class StrgUBSRAMTBShared(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 tb_height=2):

        super().__init__("strg_ub_sram_tb_shared")

        ##################################################################################
        # Capture constructor parameter...
        ##################################################################################
        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.tb_height = tb_height
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.config_width = config_width
        self.data_width = data_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support

        self.default_iterator_support = 6
        self.default_config_width = 16
        self.sram_iterator_support = 6
        self.agg_rd_addr_gen_width = 8

        ##################################################################################
        # IO
        ##################################################################################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._cycle_count = self.input("cycle_count", 16)

        self._loops_sram2tb_mux_sel = self.output("loops_sram2tb_mux_sel",
                                                  width=max(clog2(self.default_iterator_support), 1),
                                                  size=self.interconnect_output_ports,
                                                  explicit_array=True,
                                                  packed=True)

        self._loops_sram2tb_restart = self.output("loops_sram2tb_restart",
                                                  width=1,
                                                  size=self.interconnect_output_ports,
                                                  explicit_array=True,
                                                  packed=True)

        self._t_read_out = self.output("t_read_out", self.interconnect_output_ports)
        self._t_read = self.var("t_read", self.interconnect_output_ports)
        self.wire(self._t_read_out, self._t_read)

        ##################################################################################
        # TB PATHS
        ##################################################################################
        for i in range(self.interconnect_output_ports):

            # for loop for sram reads, tb writes
            loops_sram2tb = ForLoop(iterator_support=self.default_iterator_support,
                                    config_width=self.default_config_width)

            self.add_child(f"loops_buf2out_autovec_read_{i}",
                           loops_sram2tb,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._t_read[i])

            safe_wire(gen=self, w_to=self._loops_sram2tb_mux_sel[i], w_from=loops_sram2tb.ports.mux_sel_out)
            self.wire(self._loops_sram2tb_restart[i], loops_sram2tb.ports.restart)

            # sram read schedule, delay by 1 clock cycle for tb write schedule (done in tb_only)
            self.add_child(f"output_sched_gen_{i}",
                           SchedGen(iterator_support=self.default_iterator_support,
                                    # config_width=self.default_config_width),
                                    config_width=16),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           cycle_count=self._cycle_count,
                           mux_sel=loops_sram2tb.ports.mux_sel_out,
                           finished=loops_sram2tb.ports.restart,
                           valid_output=self._t_read[i])


if __name__ == "__main__":
    lake_dut = StrgUBSRAMTBShared()
    verilog(lake_dut, filename="strg_ub_sram_tb_shared.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



import kratos
from lake.attributes.formal_attr import *


def add_attrs(port, new_port, new_port_name):
    attrs = port.find_attribute(lambda a: isinstance(a, FormalAttr))
    for attr in attrs:
        attr.port_name = new_port_name
        new_port.add_attribute(attr)


def cut_generator(gen: kratos.Generator, suffix="_top"):
    # cut the generator from it's parent scope
    parent_gen = gen.internal_generator.parent_generator()
    assert parent_gen is not None, f"{gen.name} does not have a parent"
    new_ports = set()
    for port_name in gen.ports:
        port = gen.ports[port_name]
        # only need to worry about the output
        if port.port_direction == kratos.PortDirection.In.value:
            continue

        new_port_name = port_name + suffix
        new_port = parent_gen.port(port, new_port_name)

        add_attrs(port, new_port, new_port_name)

        # flip the port direction
        new_port.port_direction = kratos.PortDirection.In.value

        new_port.move_sink_to(port, new_port, parent_gen, True)
        new_ports.add(new_port.name)
    # remove from parent
    parent_gen.remove_child_generator(gen.internal_generator)
    # lifting things up
    while parent_gen.parent_generator() is not None:
        p_gen = parent_gen.parent_generator()
        for port_name in new_ports:
            port = parent_gen.get_port(port_name)
            new_port = p_gen.port(port, port_name)
            p_gen.add_stmt(port.assign(new_port))
            add_attrs(port, new_port, port_name)
        parent_gen = p_gen



from lake.models.model import Model
from lake.models.tb_model import TBModel


# transpose buffer model
class TBAModel(Model):

    def __init__(self,
                 word_width,
                 fetch_width,
                 num_tb,
                 tb_height,
                 max_range,
                 max_range_inner):

        # generation parameters
        self.word_width = word_width
        self.fetch_width = fetch_width
        self.num_tb = num_tb
        self.tb_height = tb_height
        self.max_range = max_range
        self.max_range_inner = max_range_inner

        # configuration registers
        self.config = {}
        self.config["range_outer"] = 1
        self.config["range_inner"] = 1
        self.config["stride"] = 1
        self.config["indices"] = [0]
        self.config["dimensionality"] = 1
        self.config["starting_addr"] = 0

        self.output_valid_all = []
        for i in range(self.num_tb):
            self.output_valid_all.append(0)

        self.tbs = []
        for i in range(self.num_tb):
            self.tbs.append(TBModel(self.word_width,
                                    self.fetch_width,
                                    self.num_tb,
                                    self.tb_height,
                                    self.max_range,
                                    self.max_range_inner))
            self.tbs[i].set_config(self.config)

        self.arbiter_rdy_all = []
        for i in range(self.num_tb):
            self.arbiter_rdy_all.append(0)

        self.tb_to_interconnect_data = []
        for i in range(self.tb_height):
            self.tb_to_interconnect_data.append(0)

        self.tb_to_interconnect_valid = 0
        self.tb_arbiter_rdy = 0

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

        self.output_valid_all = []
        for i in range(self.num_tb):
            self.output_valid_all.append(0)

        self.tbs = []
        for i in range(self.num_tb):
            self.tbs.append(TBModel(self.word_width,
                                    self.fetch_width,
                                    self.num_tb,
                                    self.tb_height,
                                    self.max_range,
                                    self.max_range_inner))
            self.tbs[i].set_config(self.config)

        self.arbiter_rdy_all = []
        for i in range(self.num_tb):
            self.arbiter_rdy_all.append(0)

        self.tb_to_interconnect_data = []
        for i in range(self.tb_height):
            self.tb_to_interconnect_data.append(0)

        self.tb_to_interconnect_valid = 0
        self.tb_arbiter_rdy = 0

    def set_tb_outputs(self):
        for i in range(self.num_tb):
            self.output_valid_all[i] = self.tbs[i].get_output_valid()
            if self.output_valid_all[i] == 1:
                self.tb_to_interconnect_data = self.tbs[i].get_col_pixels()
        valid_count = sum(self.output_valid_all)
        if valid_count > 0:
            self.tb_to_interconnect_valid = 1
        else:
            self.tb_to_interconnect_valid = 0

    def send_tba_rdy(self):
        for i in range(self.num_tb):
            self.arbiter_rdy_all[i] = self.tbs[i].get_rdy_to_arbiter()
        rdy_count = sum(self.arbiter_rdy_all)
        if rdy_count > 0:
            self.tb_arbiter_rdy = 1
        else:
            self.tb_arbiter_rdy = 0

    def get_ready(self):
        for i in range(self.num_tb):
            # return self.tb_arbiter_rdy
            return self.tbs[i].get_rdy_to_arbiter()

    def print_tba(self):
        print("output valid all ", self.output_valid_all)
        print("arbiter rdy all ", self.arbiter_rdy_all)
        print("tb arbiter rdy ", self.tb_arbiter_rdy)

    def print_tba_tb(self, input_data, valid_data, ack_in, tb_index_for_data, ren):
        self.tbs[0].print_tb(input_data, valid_data, ack_in, ren)
        print()

    def tba_main(self, input_data, valid_data, ack_in, tb_index_for_data, ren, mem_valid_data):
        ret_data = 0
        ret_valid = 0
        ret_rdy = 0
        for i in range(self.num_tb):
            if i == tb_index_for_data:
                valid_data_i = valid_data
                ack_in_i = ack_in
            else:
                valid_data_i = 0
                ack_in_i = 0

            (ret_data, ret_valid, ret_rdy) = self.tbs[i].interact(
                input_data, valid_data_i, ack_in_i, ren, mem_valid_data)

        # self.set_tb_outputs()
        # self.send_tba_rdy()
        # self.print_tba()

        return ret_data, ret_valid



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import *
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.utils.util import safe_wire, add_counter, decode
import kratos as kts


class StrgUBAggOnly(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 tb_height=2):

        super().__init__("strg_ub_agg_only")

        ##################################################################################
        # Capture constructor parameter...
        ##################################################################################
        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.tb_height = tb_height
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.config_width = config_width
        self.data_width = data_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support

        self.default_iterator_support = 6
        self.default_config_width = 16
        self.sram_iterator_support = 6
        self.agg_rd_addr_gen_width = 8

        ##################################################################################
        # IO
        ##################################################################################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._cycle_count = self.input("cycle_count", 16)

        self._data_in = self.input("data_in", self.data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        self._agg_read = self.input("agg_read", self.interconnect_input_ports)

        self._floop_mux_sel = self.input("floop_mux_sel",
                                         width=max(clog2(self.default_iterator_support), 1),
                                         size=self.interconnect_input_ports,
                                         explicit_array=True,
                                         packed=True)

        self._floop_restart = self.input("floop_restart",
                                         width=1,
                                         size=self.interconnect_input_ports,
                                         explicit_array=True,
                                         packed=True)

        self._agg_data_out = self.output(f"agg_data_out", self.data_width,
                                         size=(self.interconnect_input_ports,
                                               self.fetch_width),
                                         packed=True,
                                         explicit_array=True)
        self._agg_data_out.add_attribute(FormalAttr(self._agg_data_out.name, FormalSignalConstraint.SEQUENCE, "sram"))

        ##################################################################################
        # AGG RELEVANT SIGNALS
        ##################################################################################
        # Create an input to agg write scheduler + addressor for each input
        # Also need an addressor for the mux in addition to the read addr
        self._agg = self.var(f"agg",
                             width=self.data_width,
                             size=(self.interconnect_input_ports,
                                   self.agg_height,
                                   self.fetch_width),
                             packed=True,
                             explicit_array=True)

        self._agg_write = self.var("agg_write", self.interconnect_input_ports)
        # Make this based on the size
        self._agg_write_addr = self.var("agg_write_addr", 2 + clog2(self.agg_height),
                                        size=self.interconnect_input_ports,
                                        packed=True,
                                        explicit_array=True)
        self._agg_read_addr = self.var("agg_read_addr", max(1, clog2(self.agg_height)),
                                       size=self.interconnect_input_ports,
                                       packed=True,
                                       explicit_array=True)
        self._agg_read_addr_gen_out = self.var("agg_read_addr_gen_out", self.agg_rd_addr_gen_width,
                                               size=self.interconnect_input_ports,
                                               packed=True,
                                               explicit_array=True)

        ##################################################################################
        # AGG PATHS
        ##################################################################################
        for i in range(self.interconnect_input_ports):

            self.agg_iter_support = 6
            self.agg_addr_width = 4
            self.agg_range_width = 16

            forloop_ctr = ForLoop(iterator_support=self.agg_iter_support,
                                  # config_width=self.default_config_width)
                                  config_width=self.agg_range_width)
            loop_itr = forloop_ctr.get_iter()
            loop_wth = forloop_ctr.get_cfg_width()

            self.add_child(f"loops_in2buf_{i}",
                           forloop_ctr,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_write[i])

            newAG = AddrGen(iterator_support=self.agg_iter_support,
                            config_width=self.agg_addr_width)
            self.add_child(f"agg_write_addr_gen_{i}",
                           newAG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_write[i],
                           mux_sel=forloop_ctr.ports.mux_sel_out,
                           restart=forloop_ctr.ports.restart)
            safe_wire(gen=self, w_to=self._agg_write_addr[i], w_from=newAG.ports.addr_out)

            newSG = SchedGen(iterator_support=self.agg_iter_support,
                             # config_width=self.agg_addr_width)
                             config_width=16)

            self.add_child(f"agg_write_sched_gen_{i}",
                           newSG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           mux_sel=forloop_ctr.ports.mux_sel_out,
                           finished=forloop_ctr.ports.restart,
                           cycle_count=self._cycle_count,
                           valid_output=self._agg_write[i])

            @always_ff((posedge, "clk"))
            def agg_ctrl():
                if self._agg_write[i]:
                    if self.agg_height == 1:
                        self._agg[i][0][self._agg_write_addr[i][clog2(self.fetch_width) - 1, 0]] = self._data_in[i]
                    else:
                        self._agg[i][self._agg_write_addr[i]
                                     [self._agg_write_addr[0].width - 1, clog2(self.fetch_width)]]\
                                    [self._agg_write_addr[i][clog2(self.fetch_width) - 1, 0]] = self._data_in[i]

            self.add_code(agg_ctrl)

            newAG = AddrGen(iterator_support=self.default_iterator_support,
                            config_width=self.agg_addr_width)

            self.add_child(f"agg_read_addr_gen_{i}",
                           newAG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_read[i],
                           #  (self._input_port_sel_addr == const(i, self._input_port_sel_addr.width))),
                           # mux_sel=self._floop_mux_sel[i],
                           restart=self._floop_restart[i])
            safe_wire(gen=self, w_to=newAG.ports.mux_sel, w_from=self._floop_mux_sel[i])
            safe_wire(gen=self, w_to=self._agg_read_addr_gen_out[i], w_from=newAG.ports.addr_out)
            self.wire(self._agg_read_addr[i], self._agg_read_addr_gen_out[i][self._agg_read_addr.width - 1, 0])

            # Now pick out the data from the agg...
            @always_comb
            def get_agg_data():
                self._agg_data_out[i] = self._agg[i][self._agg_read_addr[i]]
            self.add_code(get_agg_data)


if __name__ == "__main__":
    lake_dut = StrgUBAggOnly()
    verilog(lake_dut, filename="strg_ub_agg_only.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
import kratos as kts
from lake.attributes.config_reg_attr import ConfigRegAttr


class SyncGroups(Generator):
    '''
    Demux the SRAM reads into writes to the transpose buffers
    '''
    def __init__(self,
                 fetch_width,
                 data_width,
                 int_out_ports):

        assert not (fetch_width & (fetch_width - 1)), "Memory width needs to be a power of 2"

        super().__init__("sync_groups")
        # Absorb inputs
        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.int_out_ports = int_out_ports
        self.groups = self.int_out_ports

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        self._ack_in = self.input("ack_in",
                                  self.int_out_ports)

        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=(self.int_out_ports,
                                         self.fw_int),
                                   explicit_array=True,
                                   packed=True)

        self._mem_valid_data = self.input("mem_valid_data",
                                          self.int_out_ports)

        self._mem_valid_data_out = self.output("mem_valid_data_out",
                                               self.int_out_ports)

        self._valid_in = self.input("valid_in",
                                    self.int_out_ports)
        # Indicates which port belongs to which synchronization group
        self._sync_group = self.input("sync_group",
                                      self.int_out_ports,
                                      size=self.int_out_ports,
                                      explicit_array=True,
                                      packed=True)
        sync_config = ConfigRegAttr("This array of one hot vectors" +
                                    " is used to denote which ports are synchronized to eachother." +
                                    " If multiple ports should output data relative to eachother" +
                                    " one should put them in the same group.")
        self._sync_group.add_attribute(sync_config)

        # Outputs
        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=(self.int_out_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        self._valid_out = self.output("valid_out",
                                      self.int_out_ports)

        # Locals
        self._sync_agg = self.var("sync_agg",
                                  self.int_out_ports,
                                  size=self.int_out_ports,
                                  explicit_array=True,
                                  packed=True)

        self._sync_valid = self.var("sync_valid", self.int_out_ports)
        self._data_reg = self.var("data_reg",
                                  self.data_width,
                                  size=(self.int_out_ports,
                                        self.fw_int),
                                  explicit_array=True,
                                  packed=True)

        self._valid_reg = self.var("valid_reg",
                                   self.int_out_ports)

        # This signal allows us to orchestrate the synchronization groups
        # at the output address controller
        self._ren_in = self.input("ren_in",
                                  self.int_out_ports)

        self._ren_int = self.var("ren_int",
                                 self.int_out_ports)

        self._rd_sync_gate = self.output("rd_sync_gate",
                                         self.int_out_ports)

        self._local_gate_bus = self.var("local_gate_bus",
                                        self.int_out_ports,
                                        size=self.int_out_ports,
                                        explicit_array=True,
                                        packed=True)

        self._local_gate_bus_n = self.var("local_gate_bus_n",
                                          self.int_out_ports,
                                          size=self.int_out_ports,
                                          explicit_array=True,
                                          packed=True)

        self.wire(self._local_gate_bus, ~self._local_gate_bus_n)

        self._local_gate_bus_tpose = self.var("local_gate_bus_tpose",
                                              self.int_out_ports,
                                              size=self.int_out_ports,
                                              explicit_array=True,
                                              packed=True)

        self._local_gate_reduced = self.var("local_gate_reduced",
                                            self.int_out_ports)

        self._local_gate_mask = self.var("local_gate_mask",
                                         self.int_out_ports,
                                         size=self.int_out_ports,
                                         explicit_array=True,
                                         packed=True)

        self._group_finished = self.var("group_finished",
                                        self.int_out_ports)

        self._grp_fin_large = self.var("grp_fin_large",
                                       self.int_out_ports,
                                       size=self.int_out_ports,
                                       explicit_array=True,
                                       packed=True)

        self._done = self.var("done", self.int_out_ports)
        self._done_alt = self.var("done_alt", self.int_out_ports)

        # Output data is ungated
        self.wire(self._data_out, self._data_reg)
        self.wire(self._rd_sync_gate, self._local_gate_reduced)
        # Valid requires gating based on sync_valid
        self.wire(self._ren_int, self._ren_in & self._local_gate_reduced)
        # Add Code
        self.add_code(self.set_sync_agg, unroll_for=True)
        self.add_code(self.set_sync_valid)
        for i in range(self.int_out_ports):
            self.add_code(self.set_sync_stage, idx=i)
        self.add_code(self.set_out_valid)
        self.add_code(self.set_reduce_gate)
        for i in range(self.groups):
            self.add_code(self.set_rd_gates, idx=i)
        self.add_code(self.set_tpose)
        self.add_code(self.set_finished)
        self.add_code(self.next_gate_mask, unroll_for=True)
        self.add_code(self.set_grp_fin, unroll_for=True)

    @always_comb
    def set_sync_agg(self):
        # For each bit, AND the valids of each group together
        # For bits not in the group, set to 1 in the corresponding mask
        # For the group
        for i in range(self.groups):
            # For the port
            for j in range(self.int_out_ports):
                if(self._sync_group[j] == (1 << i)):
                    self._sync_agg[i][j] = self._valid_reg[j]
                else:
                    self._sync_agg[i][j] = 1

    @always_comb
    def set_sync_valid(self):
        for i in range(self.int_out_ports):
            self._sync_valid[i] = self._sync_agg[i].r_and()

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_sync_stage(self, idx):
        # Ports
        if ~self._rst_n:
            self._data_reg[idx] = 0
            self._valid_reg[idx] = 0
            self._mem_valid_data_out[idx] = 0
        # Absorb input data if the whole group is valid
        elif (self._sync_valid & self._sync_group[idx]).r_or():
            self._data_reg[idx] = self._data_in[idx]
            self._mem_valid_data_out[idx] = self._mem_valid_data[idx]
            self._valid_reg[idx] = self._valid_in[idx]
        # Also absorb input data if not currently holding a valid
        elif ~self._valid_reg[idx]:
            self._data_reg[idx] = self._data_in[idx]
            self._mem_valid_data_out[idx] = self._mem_valid_data[idx]
            self._valid_reg[idx] = self._valid_in[idx]

    @always_comb
    def set_out_valid(self):
        for i in range(self.int_out_ports):
            # For each port, check its output valid
            # should just be the sync_valid
            self._valid_out[i] = (self._sync_valid & self._sync_group[i]).r_or()

    @always_comb
    def set_reduce_gate(self):
        # This time we mean port
        for i in range(self.int_out_ports):
            self._local_gate_reduced[i] = self._local_gate_bus_tpose[i].r_and()

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_rd_gates(self, idx):
        if ~self._rst_n:
            self._local_gate_bus_n[idx] = const(0, self.int_out_ports)
        elif self._group_finished[idx]:
            self._local_gate_bus_n[idx] = const(0, self.int_out_ports)
        # Bring this down eventually
        else:
            self._local_gate_bus_n[idx] = ~(self._local_gate_bus[idx] & self._local_gate_mask[idx])

    @always_comb
    def set_tpose(self):
        for i in range(self.int_out_ports):
            for j in range(self.groups):
                self._local_gate_bus_tpose[i][j] = self._local_gate_bus[j][i]

    @always_comb
    def set_grp_fin(self):
        # Group
        for i in range(self.groups):
            # Port
            for j in range(self.int_out_ports):
                self._grp_fin_large[i][j] = 1
                if self._sync_group[j] == (1 << i):
                    # If either the read had completed or is now being completed...
                    self._grp_fin_large[i][j] = (~self._local_gate_bus[i][j] |
                                                 ~self._local_gate_mask[i][j])

    @always_comb
    def set_finished(self):
        # Mark each group as done by reduce anding its wires
        for i in range(self.groups):
            self._group_finished[i] = self._grp_fin_large[i].r_and()

    @always_comb
    def next_gate_mask(self):
        for i in range(self.groups):
            for j in range(self.int_out_ports):
                self._local_gate_mask[i][j] = 1
                if (self._sync_group[j] == (1 << i)):
                    self._local_gate_mask[i][j] = ~(self._ren_int[j] & self._ack_in[j])


if __name__ == "__main__":
    db_dut = SyncGroups(fetch_width=64,
                        data_width=16,
                        int_out_ports=2)
    verilog(db_dut, filename="sync_groups.sv", optimize_if=False, check_flip_flop_always_ff=False)



from kratos import *
from lake.modules.sram_stub import SRAMStub


# Should we do this with registers or SRAM?
# What options do we have?
class VirtualRemapTable(Generator):
    def __init__(self, macro_width, logical_banks, macro_depth):
        super().__init__("virtual_remap_table")
        self._virt_addr = self.input("i_virt_addr", width)
        self._ren = self.input("i_ren", 1)
        self._out = self.output("o_data_out", width)
        self._in = self.input("i_data_in", width)
        self._clk = self.clock("i_clk")
        self._rst_n = self.reset("i_rst_n")

        # First wrap sram_stub
        sram_stub = SRAMStub(width, 1024)
        self.add_child_generator(f"u_sram_stub_0", sram_stub)
        self.wire(sram_stub.i_data, self._in)
        self.wire(self._out, sram_stub.o_data)

        self.wire(sram_stub.i_addr, 0)
        self.wire(sram_stub.i_cen, 0)
        self.wire(sram_stub.i_wen, 0)
        self.wire(sram_stub.i_clk, self._clk)
        self.wire(sram_stub.i_rst_n, self._rst_n)


if __name__ == "__main__":
    vrt_dut = VirtualRemapTable()
    verilog(vrt_dut, filename="virtual_remap_table.sv", check_active_high=False)



import kratos as kts
from enum import Enum


class ControlSignalAttr(kts.Attribute):
    def __init__(self,
                 is_control=False,
                 ignore=False,
                 doc_string=""):
        super().__init__()
        self.value = "control_signal"
        self.is_control = is_control
        self.ignore = ignore
        self.documentation = doc_string

    def set_documentation(self, new_doc):
        self.documentation = new_doc

    def get_documentation(self):
        return self.documentation

    def get_control(self):
        return self.is_control

    def get_ignore(self):
        return self.ignore



from lake.attributes.formal_attr import *
import os
from kratos import *
from lake.modules.passthru import *
from lake.modules.sram import SRAM
from lake.modules.strg_ub_vec import StrgUBVec
from lake.modules.strg_ub_thin import StrgUBThin
from lake.modules.storage_config_seq import StorageConfigSeq
from lake.modules.register_file import RegisterFile
from lake.modules.strg_fifo import StrgFIFO
from lake.modules.strg_RAM import StrgRAM
from lake.modules.chain_accessor import ChainAccessor
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.control_signal_attr import ControlSignalAttr
from lake.passes.passes import lift_config_reg, change_sram_port_names
from lake.passes.cut_generator import cut_generator
from lake.utils.sram_macro import SRAMMacroInfo
from lake.utils.util import trim_config_list, extract_formal_annotation
from lake.utils.parse_clkwork_config import map_controller, extract_controller
from lake.utils.parse_clkwork_config import extract_controller_json
from lake.modules.for_loop import ForLoop
from lake.modules.spec.sched_gen import SchedGen
import kratos as kts
from _kratos import create_wrapper_flatten
import argparse


class LakeTop(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_iterator_support=6,  # Addr Controllers
                 output_iterator_support=6,
                 input_config_width=16,
                 output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 use_sram_stub=True,
                 sram_macro_info=SRAMMacroInfo("tsmc_name"),
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 config_data_width=32,
                 config_addr_width=8,
                 num_tiles=1,
                 fifo_mode=True,
                 add_clk_enable=True,
                 add_flush=True,
                 name="LakeTop",
                 gen_addr=True,
                 stencil_valid=True,
                 formal_module=None,
                 do_config_lift=True):
        super().__init__(name, debug=True)

        self.data_width = data_width
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.banks = banks
        self.input_iterator_support = input_iterator_support
        self.output_iterator_support = output_iterator_support
        self.input_config_width = input_config_width
        self.output_config_width = output_config_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.mem_input_ports = mem_input_ports
        self.mem_output_ports = mem_output_ports
        self.use_sram_stub = use_sram_stub
        self.sram_macro_info = sram_macro_info
        self.agg_height = agg_height
        self.input_port_sched_width = clog2(self.interconnect_input_ports)
        assert self.mem_width >= self.data_width, "Data width needs to be smaller than mem"
        self.fw_int = int(self.mem_width / self.data_width)
        self.config_data_width = config_data_width
        self.config_addr_width = config_addr_width
        self.num_tiles = num_tiles
        self.read_delay = read_delay
        self.rw_same_cycle = rw_same_cycle
        self.fifo_mode = fifo_mode
        self.gen_addr = gen_addr
        self.stencil_valid = stencil_valid
        self.formal_module = formal_module

        self.data_words_per_set = 2 ** self.config_addr_width
        self.sets = int((self.fw_int * self.mem_depth) / self.data_words_per_set)

        self.sets_per_macro = max(1, int(self.mem_depth / self.data_words_per_set))
        self.total_sets = max(1, self.banks * self.sets_per_macro)

        # phases = [] TODO

        # CLK and RST
        self._clk = self.clock("clk")
        self._clk.add_attribute(FormalAttr(self._clk.name, FormalSignalConstraint.CLK))
        self._rst_n = self.reset("rst_n")
        self._rst_n.add_attribute(FormalAttr(self._rst_n.name, FormalSignalConstraint.RSTN))

        # chain data in
        self._chain_data_in = self.input("chain_data_in",
                                         self.data_width,
                                         size=self.interconnect_output_ports,
                                         packed=True,
                                         explicit_array=True)
        self._chain_data_in.add_attribute(ControlSignalAttr(False))
        self._chain_data_in.add_attribute(FormalAttr(self._chain_data_in.name, FormalSignalConstraint.SET0))

        # MAIN Inputs
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)
        self._data_in.add_attribute(ControlSignalAttr(False))
        self._data_in.add_attribute(FormalAttr(self._data_in.name, FormalSignalConstraint.SEQUENCE, "agg"))

        if self.rw_same_cycle:
            self._wr_addr_in = self.input("waddr",
                                          self.data_width,
                                          size=self.interconnect_input_ports,
                                          packed=True,
                                          explicit_array=True)
            self._wr_addr_in.add_attribute(ControlSignalAttr(False))
            self._wr_addr_in.add_attribute(FormalAttr(self._wr_addr_in.name, FormalSignalConstraint.SET0))
            self._rd_addr_in = self.input("raddr",
                                          self.data_width,
                                          size=self.interconnect_input_ports,
                                          packed=True,
                                          explicit_array=True)
            self._rd_addr_in.add_attribute(ControlSignalAttr(False))
            self._rd_addr_in.add_attribute(FormalAttr(self._rd_addr_in.name, FormalSignalConstraint.SET0))
        else:
            self._addr_in = self.input("addr_in",
                                       self.data_width,
                                       size=self.interconnect_input_ports,
                                       packed=True,
                                       explicit_array=True)
            self._addr_in.add_attribute(ControlSignalAttr(False))
            self._addr_in.add_attribute(FormalAttr(self._addr_in.name, FormalSignalConstraint.SET0))

        self._wen = self.input("wen_in", self.interconnect_input_ports)
        self._wen.add_attribute(ControlSignalAttr(True))
        self._wen.add_attribute(FormalAttr(self._wen.name, FormalSignalConstraint.SET0))

        self._ren = self.input("ren_in", self.interconnect_output_ports)
        self._ren.add_attribute(ControlSignalAttr(True))
        self._ren.add_attribute(FormalAttr(self._ren.name, FormalSignalConstraint.SET0))

        self._config_data_in = self.input("config_data_in",
                                          self.config_data_width)
        self._config_data_in.add_attribute(ControlSignalAttr(False))
        self._config_data_in.add_attribute(FormalAttr(self._ren.name, FormalSignalConstraint.SET0))

        self._config_data_in_shrt = self.var("config_data_in_shrt",
                                             self.data_width)

        self.wire(self._config_data_in_shrt, self._config_data_in[self.data_width - 1, 0])

        self._cycle_count = self.var("cycle_count", 16)
        self.add_code(self.cycle_count_inc)

        if self.stencil_valid:

            self.stencil_valid_width = 16

            self._stencil_valid = self.output("stencil_valid", 1)
            self._stencil_valid.add_attribute(ControlSignalAttr(False))
            self._loops_stencil_valid = ForLoop(iterator_support=6,
                                                config_width=16)
            self._stencil_valid_int = self.var("stencil_valid_internal", 1)

            # Loop Iterators for stencil valid...
            self.add_child(f"loops_stencil_valid",
                           self._loops_stencil_valid,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._stencil_valid_int)
            # Schedule Generator for stencil valid...
            self.add_child(f"stencil_valid_sched_gen",
                           SchedGen(iterator_support=6,
                                    config_width=16),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           cycle_count=self._cycle_count,
                           mux_sel=self._loops_stencil_valid.ports.mux_sel_out,
                           finished=self._loops_stencil_valid.ports.restart,
                           valid_output=self._stencil_valid_int)
            # Wire out internal wire
            # self.wire(self._stencil_valid, self._stencil_valid_int & self._stencil_valid_gate & self._use_stencil_valid)
            self.wire(self._stencil_valid, self._stencil_valid_int)

        self._config_addr_in = self.input("config_addr_in",
                                          self.config_addr_width)
        self._config_addr_in.add_attribute(ControlSignalAttr(False))

        self._config_data_out_shrt = self.var("config_data_out_shrt", self.data_width,
                                              size=self.total_sets,
                                              explicit_array=True,
                                              packed=True)

        self._config_data_out = self.output("config_data_out", self.config_data_width,
                                            size=self.total_sets,
                                            explicit_array=True,
                                            packed=True)
        self._config_data_out.add_attribute(ControlSignalAttr(False))

        self._clk_en = self.clock_en("clk_en", 1)

        for i in range(self.total_sets):
            self.wire(self._config_data_out[i],
                      self._config_data_out_shrt[i].extend(self.config_data_width))

        self._config_read = self.input("config_read", 1)
        self._config_read.add_attribute(ControlSignalAttr(False))

        self._config_write = self.input("config_write", 1)
        self._config_write.add_attribute(ControlSignalAttr(False))

        self._config_en = self.input("config_en", self.total_sets)
        self._config_en.add_attribute(ControlSignalAttr(False))

        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)
        self._data_out.add_attribute(ControlSignalAttr(False))
        self._data_out.add_attribute(FormalAttr(self._data_out.name, FormalSignalConstraint.SEQUENCE, "tb"))

        # self._valid_out = self.output("valid_out",
        #                               self.interconnect_output_ports)
        # self._valid_out.add_attribute(ControlSignalAttr(False))

        self._data_out_tile = self.var("data_out_tile",
                                       self.data_width,
                                       size=self.interconnect_output_ports,
                                       packed=True,
                                       explicit_array=True)

        self._valid_out_tile = self.output("valid_out",
                                           self.interconnect_output_ports)
        self._valid_out_tile.add_attribute(ControlSignalAttr(False))

        self.address_width = clog2(self.num_tiles * self.mem_depth)

        # Add tile enable!
        self._tile_en = self.input("tile_en", 1)
        self._tile_en.add_attribute(ConfigRegAttr("Tile logic enable manifested as clock gate"))
        self._tile_en.add_attribute(FormalAttr(self._tile_en.name, FormalSignalConstraint.SET1))

        # either normal or fifo mode rn...
        self.num_modes = 3
        self._mode = self.input("mode", max(1, clog2(self.num_modes)))
        self._mode.add_attribute(ConfigRegAttr("MODE!"))
        self._mode.add_attribute(FormalAttr(self._mode.name, FormalSignalConstraint.SET0))

        # Currenlt mode = 0 is UB, mode = 1 is FIFO
        gclk = self.var("gclk", 1)
        self._gclk = kts.util.clock(gclk)
        self.wire(gclk, kts.util.clock(self._clk & self._tile_en))

        self._mem_data_out = self.var("mem_data_out",
                                      self.data_width,
                                      size=(self.banks,
                                            self.mem_output_ports,
                                            self.fw_int),
                                      packed=True,
                                      explicit_array=True)

        if self.formal_module == "sram":
            self._formal_mem_data = self.output("formal_mem_data",
                                                self.data_width,
                                                size=(self.banks,
                                                      self.mem_output_ports,
                                                      self.fw_int),
                                                packed=True,
                                                explicit_array=True)
            self._formal_mem_data.add_attribute(
                FormalAttr(self._formal_mem_data.name, FormalSignalConstraint.SEQUENCE, self.formal_module))

            self.wire(self._formal_mem_data, self._mem_data_out)
        elif formal_module == "tb":
            self._formal_mem_data = self.input("formal_mem_data",
                                               self.data_width,
                                               size=(self.banks,
                                                     self.mem_output_ports,
                                                     self.fw_int),
                                               packed=True,
                                               explicit_array=True)
            self._formal_mem_data.add_attribute(
                FormalAttr(self._formal_mem_data.name, FormalSignalConstraint.SEQUENCE, self.formal_module))

            self.wire(self._mem_data_out, self._formal_mem_data)

        self._mem_data_low_pt = self.var("mem_data_low_pt",
                                         self.data_width,
                                         size=(self.banks,
                                               self.fw_int),
                                         packed=True,
                                         explicit_array=True)

        for i in range(self.banks):
            self.wire(self._mem_data_low_pt[i], self._mem_data_out[i][0])

        self._mem_data_in = self.var("mem_data_in",
                                     self.data_width,
                                     size=(self.banks,
                                           self.mem_input_ports,
                                           self.fw_int),
                                     packed=True,
                                     explicit_array=True)

        self._mem_data_dp = self.var("mem_data_dp",
                                     self.data_width,
                                     size=(self.banks,
                                           self.mem_input_ports,
                                           self.fw_int),
                                     packed=True,
                                     explicit_array=True)

        self._mem_data_cfg = self.var("mem_data_cfg",
                                      self.data_width,
                                      size=self.fw_int,
                                      packed=True,
                                      explicit_array=True)

        if self.rw_same_cycle:
            self._wr_mem_addr_dp = self.var("wr_mem_addr_dp",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_input_ports),
                                            explicit_array=True,
                                            packed=True)
            self._wr_mem_addr_in = self.var("wr_mem_addr_in",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_input_ports),
                                            explicit_array=True,
                                            packed=True)
            self._rd_mem_addr_dp = self.var("rd_mem_addr_dp",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_output_ports),
                                            explicit_array=True,
                                            packed=True)
            self._rd_mem_addr_in = self.var("rd_mem_addr_in",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_output_ports),
                                            explicit_array=True,
                                            packed=True)

        else:
            self._mem_addr_dp = self.var("mem_addr_dp",
                                         self.address_width,
                                         size=(self.banks,
                                               self.mem_input_ports),
                                         packed=True,
                                         explicit_array=True)
            self._mem_addr_in = self.var("mem_addr_in",
                                         self.address_width,
                                         size=(self.banks,
                                               self.mem_input_ports),
                                         packed=True,
                                         explicit_array=True)

        self._mem_addr_cfg = self.var("mem_addr_cfg", self.address_width)

        self._mem_ren_cfg = self.var("mem_ren_cfg", self.banks)
        self._mem_wen_cfg = self.var("mem_wen_cfg", self.banks)

        self._mem_cen_dp = self.var("mem_cen_dp", self.mem_output_ports,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)

        self._mem_wen_dp = self.var("mem_wen_dp", self.mem_input_ports,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)
        self._mem_cen_in = self.var("mem_cen_in", self.mem_output_ports,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)

        self._mem_wen_in = self.var("mem_wen_in", self.mem_input_ports,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)

        ####################################
        ##### DEMUX WRITE/SRAM WRAPPER #####
        ####################################

        stg_cfg_seq = StorageConfigSeq(data_width=self.data_width,
                                       config_addr_width=self.config_addr_width,
                                       addr_width=self.address_width,
                                       fetch_width=self.mem_width,
                                       total_sets=self.total_sets,
                                       sets_per_macro=self.sets_per_macro)

        # The clock to config sequencer needs to be the normal clock or
        # if the tile is off, we bring the clock back in based on config_en
        cfg_seq_clk = self.var("cfg_seq_clk", 1)
        self._cfg_seq_clk = kts.util.clock(cfg_seq_clk)
        # self.wire(cfg_seq_clk, kts.util.clock(self._gclk | (self._clk & self._config_en.r_or())))
        self.wire(cfg_seq_clk, kts.util.clock(self._gclk))

        self.add_child(f"config_seq", stg_cfg_seq,
                       clk=self._cfg_seq_clk,
                       rst_n=self._rst_n,
                       clk_en=self._clk_en | self._config_en.r_or(),
                       config_data_in=self._config_data_in_shrt,
                       config_addr_in=self._config_addr_in,
                       config_wr=self._config_write,
                       config_rd=self._config_read,
                       config_en=self._config_en,
                       rd_data_stg=self._mem_data_low_pt,
                       wr_data=self._mem_data_cfg,
                       rd_data_out=self._config_data_out_shrt,
                       addr_out=self._mem_addr_cfg,
                       wen_out=self._mem_wen_cfg,
                       ren_out=self._mem_ren_cfg)

        for i in range(self.banks):
            # If we are not sharing the ports...
            if self.rw_same_cycle:
                self.wire(self._mem_wen_in[i][0], ternary(self._config_en.r_or(),
                                                          self._mem_wen_cfg[i],
                                                          self._mem_wen_dp[i][0]))
                # Treat cen as ren for all intents and purpose
                self.wire(self._mem_cen_in[i][0], ternary(self._config_en.r_or(),
                                                          self._mem_ren_cfg[i],
                                                          self._mem_cen_dp[i][0]))
                self.wire(self._wr_mem_addr_in[i][0], ternary(self._config_en.r_or(),
                                                              self._mem_addr_cfg,
                                                              self._wr_mem_addr_dp[i][0]))
                self.wire(self._rd_mem_addr_in[i][0], ternary(self._config_en.r_or(),
                                                              self._mem_addr_cfg,
                                                              self._rd_mem_addr_dp[i][0]))
                self.wire(self._mem_data_in[i][0], ternary(self._config_en.r_or(),
                                                           self._mem_data_cfg,
                                                           self._mem_data_dp[i][0]))
                # Don't route the config to any but the first port
                for j in range(self.mem_input_ports - 1):
                    self.wire(self._mem_wen_in[i][j + 1], self._mem_wen_dp[i][j + 1])
                    self.wire(self._wr_mem_addr_in[i][j + 1], self._wr_mem_addr_dp[i][j + 1])
                    self.wire(self._mem_data_in[i][j + 1], self._mem_data_dp[i][j + 1])
                for j in range(self.mem_output_ports - 1):
                    self.wire(self._mem_cen_in[i][j + 1], self._mem_cen_dp[i][j + 1])
                    self.wire(self._rd_mem_addr_in[i][j + 1], self._rd_mem_addr_dp[i][j + 1])
            else:
                self.wire(self._mem_wen_in[i][0], ternary(self._config_en.r_or(),
                                                          self._mem_wen_cfg[i],
                                                          self._mem_wen_dp[i][0]))
                self.wire(self._mem_cen_in[i][0], ternary(self._config_en.r_or(),
                                                          self._mem_wen_cfg[i] | self._mem_ren_cfg[i],
                                                          self._mem_cen_dp[i][0]))
                self.wire(self._mem_addr_in[i][0], ternary(self._config_en.r_or(),
                                                           self._mem_addr_cfg,
                                                           self._mem_addr_dp[i][0]))
                self.wire(self._mem_data_in[i][0], ternary(self._config_en.r_or(),
                                                           self._mem_data_cfg,
                                                           self._mem_data_dp[i][0]))
                # Don't route the config to any but the first port
                for j in range(self.mem_input_ports - 1):
                    self.wire(self._mem_wen_in[i][j + 1], self._mem_wen_dp[i][j + 1])
                    self.wire(self._mem_cen_in[i][j + 1], self._mem_cen_dp[i][j + 1])
                    self.wire(self._mem_addr_in[i][j + 1], self._mem_addr_dp[i][j + 1])
                    self.wire(self._mem_data_in[i][j + 1], self._mem_data_dp[i][j + 1])

        self._mem_data_ub = self.var("mem_data_ub",
                                     self.data_width,
                                     size=(self.banks,
                                           self.mem_input_ports,
                                           self.fw_int),
                                     packed=True,
                                     explicit_array=True)

        if self.fw_int > 1:
            strg_ub = StrgUBVec(data_width=self.data_width,
                                mem_width=self.mem_width,
                                mem_depth=self.mem_depth,
                                banks=self.banks,
                                input_addr_iterator_support=self.input_iterator_support,
                                output_addr_iterator_support=self.output_iterator_support,
                                input_sched_iterator_support=self.input_iterator_support,
                                output_sched_iterator_support=self.output_iterator_support,
                                interconnect_input_ports=self.interconnect_input_ports,
                                interconnect_output_ports=self.interconnect_output_ports,
                                mem_input_ports=self.mem_input_ports,
                                mem_output_ports=self.mem_output_ports,
                                read_delay=self.read_delay,
                                rw_same_cycle=self.rw_same_cycle,
                                agg_height=self.agg_height,
                                config_width=self.input_config_width,
                                agg_data_top=(self.formal_module == "agg"))

        else:

            strg_ub = StrgUBThin(data_width=self.data_width,
                                 mem_width=self.mem_width,
                                 mem_depth=self.mem_depth,
                                 banks=self.banks,
                                 input_addr_iterator_support=self.input_iterator_support,
                                 input_sched_iterator_support=self.input_iterator_support,
                                 output_addr_iterator_support=self.output_iterator_support,
                                 output_sched_iterator_support=self.output_iterator_support,
                                 interconnect_input_ports=self.interconnect_input_ports,
                                 interconnect_output_ports=self.interconnect_output_ports,
                                 config_width=self.input_config_width,
                                 mem_input_ports=self.mem_input_ports,
                                 mem_output_ports=self.mem_output_ports,
                                 read_delay=self.read_delay,
                                 rw_same_cycle=self.rw_same_cycle,
                                 gen_addr=self.gen_addr)

        self._ub_data_to_mem = self.var("ub_data_to_mem",
                                        self.data_width,
                                        size=(self.banks,
                                              self.mem_input_ports,
                                              self.fw_int),
                                        packed=True,
                                        explicit_array=True)

        self._ub_data_out = self.var("ub_data_out",
                                     self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        self._ub_valid_out = self.var("ub_valid_out",
                                      self.interconnect_output_ports)

        self._ub_wen_to_mem = self.var("ub_wen_to_mem", self.mem_output_ports,
                                       size=self.banks,
                                       explicit_array=True,
                                       packed=True)
        self._ub_cen_to_mem = self.var("ub_cen_to_mem", self.mem_input_ports,
                                       size=self.banks,
                                       explicit_array=True,
                                       packed=True)

        if self.rw_same_cycle:
            self._ub_wr_addr_to_mem = self.var("ub_wr_addr_to_mem",
                                               self.address_width,
                                               size=(self.banks,
                                                     self.mem_input_ports),
                                               explicit_array=True,
                                               packed=True)
            self._ub_rd_addr_to_mem = self.var("ub_rd_addr_to_mem",
                                               self.address_width,
                                               size=(self.banks,
                                                     self.mem_output_ports),
                                               explicit_array=True,
                                               packed=True)
        else:
            self._ub_addr_to_mem = self.var("ub_addr_to_mem",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_input_ports),
                                            packed=True,
                                            explicit_array=True)

        self._accessor_output = self.var("accessor_output", self.interconnect_output_ports)

        self.add_child("strg_ub", strg_ub,
                       # clk + rst
                       clk=self._gclk,
                       rst_n=self._rst_n,
                       # inputs
                       data_in=self._data_in,
                       data_from_strg=self._mem_data_out,
                       # outputs
                       data_out=self._ub_data_out,
                       data_to_strg=self._ub_data_to_mem,
                       wen_to_strg=self._ub_wen_to_mem,
                       accessor_output=self._accessor_output)

        if self.formal_module == "agg":
            self._formal_agg_data_out = self.output("formal_agg_data_out", self.data_width,
                                                    size=(self.interconnect_input_ports,
                                                          self.fw_int),
                                                    packed=True,
                                                    explicit_array=True)
            self._formal_agg_data_out.add_attribute(
                FormalAttr(self._formal_agg_data_out.name, FormalSignalConstraint.SEQUENCE, "agg"))

            self.wire(self._formal_agg_data_out, strg_ub.ports.strg_ub_agg_data_out)
        # Handle different names - sorry
        if self.rw_same_cycle:
            self.wire(strg_ub.ports.ren_to_strg, self._ub_cen_to_mem)
        else:
            self.wire(strg_ub.ports.cen_to_strg, self._ub_cen_to_mem)

        # Wire addrs
        if self.rw_same_cycle:
            self.wire(self._ub_wr_addr_to_mem, strg_ub.ports.wr_addr_out)
            self.wire(self._ub_rd_addr_to_mem, strg_ub.ports.rd_addr_out)
        # wire single addr
        else:
            self.wire(self._ub_addr_to_mem, strg_ub.ports.addr_out)

        if self.gen_addr is False:
            self.wire(strg_ub.ports.read_addr, self._rd_addr_in[0])
            self.wire(strg_ub.ports.write_addr, self._wr_addr_in[0])
            self.wire(strg_ub.ports.wen_in, self._wen)
            self.wire(strg_ub.ports.ren_in, self._ren)

        # Wrap sram_stub
        if self.read_delay == 1 and self.rw_same_cycle is False:

            self._all_data_to_mem = self.var("all_data_to_mem", self.data_width,
                                             size=(self.num_modes,
                                                   self.banks,
                                                   self.fw_int),
                                             explicit_array=True,
                                             packed=True)
            self._all_wen_to_mem = self.var("all_wen_to_mem", self.mem_input_ports,
                                            size=(self.num_modes,
                                                  self.banks),
                                            explicit_array=True,
                                            packed=True)
            self._all_ren_to_mem = self.var("all_ren_to_mem", self.mem_output_ports,
                                            size=(self.num_modes,
                                                  self.banks),
                                            explicit_array=True,
                                            packed=True)

            self._all_addr_to_mem = self.var("all_addr_to_mem", self.address_width,
                                             size=(self.num_modes,
                                                   self.banks),
                                             explicit_array=True,
                                             packed=True)

            self._fifo_data_out = self.var("fifo_data_out", self.data_width)
            self._fifo_valid_out = self.var("fifo_valid_out", 1)
            self._fifo_empty = self.var("fifo_empty", 1)
            self._fifo_full = self.var("fifo_full", 1)
            self._fifo_data_to_mem = self.var("fifo_data_to_mem", self.data_width,
                                              size=(self.banks,
                                                    self.fw_int),
                                              explicit_array=True,
                                              packed=True)
            self._fifo_wen_to_mem = self.var("fifo_wen_to_mem", self.banks)
            self._fifo_ren_to_mem = self.var("fifo_ren_to_mem", self.banks)

            self._fifo_addr_to_mem = self.var("fifo_addr_to_mem", self.address_width,
                                              size=self.banks,
                                              explicit_array=True,
                                              packed=True)

            self._sram_data_out = self.var("sram_data_out", self.data_width)
            self._sram_valid_out = self.var("sram_valid_out", 1)
            self._sram_empty = self.var("sram_empty", 1)
            self._sram_full = self.var("sram_full", 1)
            self._sram_data_to_mem = self.var("sram_data_to_mem", self.data_width,
                                              size=(self.banks,
                                                    self.fw_int),
                                              explicit_array=True,
                                              packed=True)
            self._sram_wen_to_mem = self.var("sram_wen_to_mem", self.banks)
            self._sram_ren_to_mem = self.var("sram_ren_to_mem", self.banks)

            self._sram_addr_to_mem = self.var("sram_addr_to_mem", self.address_width,
                                              size=self.banks,
                                              explicit_array=True,
                                              packed=True)

            if self.fw_int > 1:
                self._sram_ready_out = self.output("sram_ready_out", 1)
                self._sram_ready_out.add_attribute(ControlSignalAttr(False))

            strg_ram = StrgRAM(data_width=self.data_width,
                               banks=self.banks,
                               memory_width=self.mem_width,
                               memory_depth=self.mem_depth,
                               num_tiles=self.num_tiles,
                               rw_same_cycle=self.rw_same_cycle,
                               read_delay=self.read_delay,
                               addr_width=16,
                               prioritize_write=True)

            self.add_child("sram_ctrl", strg_ram,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           wen=self._wen[0],
                           ren=self._ren[0],
                           data_in=self._data_in[0],
                           wr_addr_in=self._addr_in[0],
                           rd_addr_in=self._addr_in[0],
                           data_from_strg=self._mem_data_out,
                           data_out=self._sram_data_out,
                           valid_out=self._sram_valid_out,
                           data_to_strg=self._sram_data_to_mem,
                           wen_to_strg=self._sram_wen_to_mem,
                           ren_to_strg=self._sram_ren_to_mem,
                           addr_out=self._sram_addr_to_mem)

            if self.fw_int > 1:
                self.wire(self._sram_ready_out, strg_ram.ports.ready)

            # If we have the fifo mode enabled -
            # Instantiate a FIFO first off...
            stfo = StrgFIFO(data_width=self.data_width,
                            banks=self.banks,
                            memory_width=self.mem_width,
                            rw_same_cycle=False,
                            read_delay=self.read_delay,
                            addr_width=self.address_width)
            self.add_child("fifo_ctrl", stfo,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           data_in=self._data_in[0],
                           push=self._wen[0],
                           pop=self._ren[0],
                           data_from_strg=self._mem_data_out,
                           data_out=self._fifo_data_out,
                           valid_out=self._fifo_valid_out,
                           empty=self._fifo_empty,
                           full=self._fifo_full,
                           data_to_strg=self._fifo_data_to_mem,
                           wen_to_strg=self._fifo_wen_to_mem,
                           ren_to_strg=self._fifo_ren_to_mem,
                           addr_out=self._fifo_addr_to_mem)

            self._empty = self.output("empty", 1)
            self._empty.add_attribute(ControlSignalAttr(False))

            self._full = self.output("full", 1)
            self._full.add_attribute(ControlSignalAttr(False))

            self.wire(self._empty, self._fifo_empty)
            self.wire(self._full, self._fifo_full)

            self.wire(self._all_data_to_mem[0], self._ub_data_to_mem)
            self.wire(self._all_wen_to_mem[0], self._ub_wen_to_mem)
            self.wire(self._all_ren_to_mem[0], self._ub_cen_to_mem)
            self.wire(self._all_addr_to_mem[0], self._ub_addr_to_mem)

            self.wire(self._all_data_to_mem[1], self._fifo_data_to_mem)
            for i in range(self.banks):
                self.wire(self._all_wen_to_mem[1][i], self._fifo_wen_to_mem[i])
                self.wire(self._all_ren_to_mem[1][i], self._fifo_ren_to_mem[i])
                self.wire(self._all_addr_to_mem[1][i], self._fifo_addr_to_mem[i])

            self.wire(self._all_data_to_mem[2], self._sram_data_to_mem)
            for i in range(self.banks):
                self.wire(self._all_wen_to_mem[2][i], self._sram_wen_to_mem[i])
                self.wire(self._all_ren_to_mem[2][i], self._sram_ren_to_mem[i])
                self.wire(self._all_addr_to_mem[2][i], self._sram_addr_to_mem[i])

            # Mux all of these signals when in FIFO mode
            self.wire(self._mem_data_dp, self._all_data_to_mem[self._mode])
            self.wire(self._mem_cen_dp, self._all_ren_to_mem[self._mode] | self._all_wen_to_mem[self._mode])
            self.wire(self._mem_wen_dp, self._all_wen_to_mem[self._mode])
            self.wire(self._mem_addr_dp, self._all_addr_to_mem[self._mode])

            # have SRAM when not doing formal module generation or when generating
            # SRAM modular problem for formal team
            if self.formal_module is None or self.formal_module == "sram":
                for i in range(self.banks):
                    mbank = SRAM(use_sram_stub=self.use_sram_stub,
                                 sram_name=self.sram_macro_info.name,
                                 data_width=self.data_width,
                                 fw_int=self.fw_int,
                                 mem_depth=self.mem_depth,
                                 mem_input_ports=self.mem_input_ports,
                                 mem_output_ports=self.mem_output_ports,
                                 address_width=self.address_width,
                                 bank_num=i,
                                 num_tiles=self.num_tiles)

                    self.add_child(f"mem_{i}", mbank,
                                   clk=self._gclk,
                                   clk_en=self._clk_en | self._config_en.r_or(),
                                   mem_data_in_bank=self._mem_data_in[i],
                                   mem_data_out_bank=self._mem_data_out[i],
                                   mem_addr_in_bank=self._mem_addr_in[i],
                                   mem_cen_in_bank=self._mem_cen_in[i],
                                   mem_wen_in_bank=self._mem_wen_in[i],
                                   wtsel=self.sram_macro_info.wtsel_value,
                                   rtsel=self.sram_macro_info.rtsel_value)
        else:

            self.wire(self._mem_data_dp, self._ub_data_to_mem)
            self.wire(self._mem_wen_dp, self._ub_wen_to_mem)
            self.wire(self._mem_cen_dp, self._ub_cen_to_mem)
            if self.rw_same_cycle:
                self.wire(self._wr_mem_addr_dp, self._ub_wr_addr_to_mem)
                self.wire(self._rd_mem_addr_dp, self._ub_rd_addr_to_mem)
            else:
                self.wire(self._mem_addr_dp, self._ub_addr_to_mem)

            for i in range(self.banks):
                rfile = RegisterFile(data_width=self.data_width,
                                     write_ports=self.mem_input_ports,
                                     read_ports=self.mem_output_ports,
                                     width_mult=self.fw_int,
                                     depth=self.mem_depth,
                                     read_delay=self.read_delay)
                if self.rw_same_cycle:
                    self.add_child(f"rf_{i}", rfile,
                                   clk=self._gclk,
                                   rst_n=self._rst_n,
                                   wen=self._mem_wen_in[i],
                                   wr_addr=self._wr_mem_addr_in[i],
                                   rd_addr=self._rd_mem_addr_in[i],
                                   data_in=self._mem_data_in[i],
                                   data_out=self._mem_data_out[i])
                else:
                    self.add_child(f"rf_{i}", rfile,
                                   clk=self._gclk,
                                   rst_n=self._rst_n,
                                   wen=self._mem_wen_in[i],
                                   wr_addr=self._mem_addr_in[i],
                                   rd_addr=self._mem_addr_in[i],
                                   data_in=self._mem_data_in[i],
                                   data_out=self._mem_data_out[i])
                if self.read_delay == 1:
                    self.wire(rfile.ports.ren, self._mem_cen_dp[0])

        if self.fifo_mode:
            self._all_data_out = self.var("all_data_out", self.data_width,
                                          size=self.num_modes,
                                          explicit_array=True,
                                          packed=True)
            self._all_valid_out = self.var("all_valid_out", self.num_modes)

            self.wire(self._all_data_out[0], self._ub_data_out[0])
            self.wire(self._all_valid_out[0], self._ub_valid_out[0])

            self.wire(self._all_data_out[1], self._fifo_data_out)
            self.wire(self._all_valid_out[1], self._fifo_valid_out)

            self.wire(self._all_data_out[2], self._sram_data_out)
            self.wire(self._all_valid_out[2], self._sram_valid_out)

            self.wire(self._data_out_tile[0], self._all_data_out[self._mode])
            self.wire(self._valid_out_tile[0], self._all_valid_out[self._mode])
        else:
            self.wire(self._data_out_tile[0], self._ub_data_out[0])
            self.wire(self._valid_out_tile[0], self._ub_valid_out[0])

        for i in range(self.interconnect_output_ports - 1):
            self.wire(self._data_out_tile[i + 1], self._ub_data_out[i + 1])
            self.wire(self._valid_out_tile[i + 1], self._ub_valid_out[i + 1])

        chaining = ChainAccessor(data_width=self.data_width,
                                 interconnect_output_ports=self.interconnect_output_ports)

        self._mode_mask = self.var("mode_mask", self._accessor_output.width)
        self.wire(self._mode_mask[0], self._mode.r_or())
        for i in range(self._accessor_output.width - 1):
            self.wire(self._mode_mask[i + 1], kts.const(0, 1))

        self.add_child(f"chain", chaining,
                       curr_tile_data_out=self._data_out_tile,
                       chain_data_in=self._chain_data_in,
                       accessor_output=(self._accessor_output | self._mode_mask),
                       data_out_tile=self._data_out)

        ########################
        ##### CLOCK ENABLE #####
        ########################
        if add_clk_enable:
            # self.clock_en("clk_en")
            kts.passes.auto_insert_clock_enable(self.internal_generator)
            # Add input attr and formal attr...
            clk_en_port = self.internal_generator.get_port("clk_en")
            clk_en_port.add_attribute(ControlSignalAttr(False))
            clk_en_port.add_attribute(FormalAttr(clk_en_port.name, FormalSignalConstraint.SET1))
        if add_flush:
            self.add_attribute("sync-reset=flush")
            kts.passes.auto_insert_sync_reset(self.internal_generator)
            flush_port = self.internal_generator.get_port("flush")
            flush_port.add_attribute(ControlSignalAttr(True))

        if do_config_lift:
            lift_config_reg(self.internal_generator)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def cycle_count_inc(self):
        if ~self._rst_n:
            self._cycle_count = 0
        else:
            self._cycle_count = self._cycle_count + 1

    def supports(self, prop):
        attr = getattr(self, prop)
        if attr:
            return attr
        else:
            return False

    def get_static_bitstream_json(self,
                                  root_node):
        # Dummy variables to fill in later when compiler
        # generates different collateral for different designs
        input_ports = 1
        output_ports = 1

        flattened = create_wrapper_flatten(self.internal_generator.clone(),
                                           self.name + "_W")

        # Store all configurations here
        config = []

        # Compiler tells us to turn on the chain enable...
        if "chain_en" in root_node:
            config.append(("chain_chain_en", 1))

        if "in2agg_0" in root_node:
            in2agg_0 = map_controller(extract_controller_json(root_node["in2agg_0"]), "in2agg_0")
            config.append(("strg_ub_agg_only_agg_write_addr_gen_0_starting_addr", in2agg_0.in_data_strt))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_0_enable", 1))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_0_sched_addr_gen_starting_addr", in2agg_0.cyc_strt))
            config.append(("strg_ub_agg_only_loops_in2buf_0_dimensionality", in2agg_0.dim))
            for i in range(in2agg_0.dim):
                config.append((f"strg_ub_agg_only_loops_in2buf_0_ranges_{i}", in2agg_0.extent[i]))
                config.append((f"strg_ub_agg_only_agg_write_addr_gen_0_strides_{i}", in2agg_0.in_data_stride[i]))
                config.append((f"strg_ub_agg_only_agg_write_sched_gen_0_sched_addr_gen_strides_{i}", in2agg_0.cyc_stride[i]))

        if "in2agg_1" in root_node:
            in2agg_1 = map_controller(extract_controller_json(root_node["in2agg_1"]), "in2agg_1")
            config.append(("strg_ub_agg_only_agg_write_addr_gen_1_starting_addr", in2agg_1.in_data_strt))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_1_enable", 1))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_1_sched_addr_gen_starting_addr", in2agg_1.cyc_strt))
            config.append(("strg_ub_agg_only_loops_in2buf_1_dimensionality", in2agg_1.dim))
            for i in range(in2agg_1.dim):
                config.append((f"strg_ub_agg_only_loops_in2buf_1_ranges_{i}", in2agg_1.extent[i]))
                config.append((f"strg_ub_agg_only_agg_write_addr_gen_1_strides_{i}", in2agg_1.in_data_stride[i]))
                config.append((f"strg_ub_agg_only_agg_write_sched_gen_1_sched_addr_gen_strides_{i}", in2agg_1.cyc_stride[i]))

        if "agg2sram_0" in root_node:
            agg2sram_0 = map_controller(extract_controller_json(root_node["agg2sram_0"]), "agg2sram_0")
            config.append(("strg_ub_agg_sram_shared_loops_in2buf_autovec_write_0_dimensionality", agg2sram_0.dim))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_0_enable", 1))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_0_sched_addr_gen_starting_addr", agg2sram_0.cyc_strt))
            config.append(("strg_ub_agg_only_agg_read_addr_gen_0_starting_addr", agg2sram_0.out_data_strt))
            config.append(("strg_ub_sram_only_input_addr_gen_0_starting_addr", agg2sram_0.in_data_strt))
            for i in range(agg2sram_0.dim):
                config.append((f"strg_ub_agg_only_agg_read_addr_gen_0_strides_{i}", agg2sram_0.out_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_loops_in2buf_autovec_write_0_ranges_{i}", agg2sram_0.extent[i]))
                config.append((f"strg_ub_sram_only_input_addr_gen_0_strides_{i}", agg2sram_0.in_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_agg_read_sched_gen_0_sched_addr_gen_strides_{i}", agg2sram_0.cyc_stride[i]))

        if "agg2sram_1" in root_node:
            agg2sram_1 = map_controller(extract_controller_json(root_node["agg2sram_1"]), "agg2sram_1")
            config.append(("strg_ub_agg_sram_shared_loops_in2buf_autovec_write_1_dimensionality", agg2sram_1.dim))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_1_enable", 1))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_1_sched_addr_gen_starting_addr", agg2sram_1.cyc_strt))
            config.append(("strg_ub_agg_only_agg_read_addr_gen_1_starting_addr", agg2sram_1.out_data_strt))
            config.append(("strg_ub_sram_only_input_addr_gen_1_starting_addr", agg2sram_1.in_data_strt))
            for i in range(agg2sram_1.dim):
                config.append((f"strg_ub_agg_only_agg_read_addr_gen_1_strides_{i}", agg2sram_1.out_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_loops_in2buf_autovec_write_1_ranges_{i}", agg2sram_1.extent[i]))
                config.append((f"strg_ub_sram_only_input_addr_gen_1_strides_{i}", agg2sram_1.in_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_agg_read_sched_gen_1_sched_addr_gen_strides_{i}", agg2sram_1.cyc_stride[i]))

        # Count tbs
        num_tbs = 0

        if "tb2out_0" in root_node:
            num_tbs += 1
            tb2out_0 = map_controller(extract_controller_json(root_node["tb2out_0"]), "tb2out_0")
            config.append(("strg_ub_tb_only_tb_read_sched_gen_0_enable", 1))
            config.append(("strg_ub_tb_only_tb_read_sched_gen_0_sched_addr_gen_starting_addr", tb2out_0.cyc_strt))
            config.append(("strg_ub_tb_only_tb_read_addr_gen_0_starting_addr", tb2out_0.out_data_strt))
            config.append(("strg_ub_tb_only_loops_buf2out_read_0_dimensionality", tb2out_0.dim))
            for i in range(tb2out_0.dim):
                config.append((f"strg_ub_tb_only_loops_buf2out_read_0_ranges_{i}", tb2out_0.extent[i]))
                config.append((f"strg_ub_tb_only_tb_read_addr_gen_0_strides_{i}", tb2out_0.out_data_stride[i]))
                config.append((f"strg_ub_tb_only_tb_read_sched_gen_0_sched_addr_gen_strides_{i}", tb2out_0.cyc_stride[i]))

        if "tb2out_1" in root_node:
            num_tbs += 1
            tb2out_1 = map_controller(extract_controller_json(root_node["tb2out_1"]), "tb2out_1")
            config.append(("strg_ub_tb_only_tb_read_sched_gen_1_enable", 1))
            config.append(("strg_ub_tb_only_tb_read_addr_gen_1_starting_addr", tb2out_1.out_data_strt))
            config.append(("strg_ub_tb_only_tb_read_sched_gen_1_sched_addr_gen_starting_addr", tb2out_1.cyc_strt))
            config.append(("strg_ub_tb_only_loops_buf2out_read_1_dimensionality", tb2out_1.dim))
            for i in range(tb2out_1.dim):
                config.append((f"strg_ub_tb_only_loops_buf2out_read_1_ranges_{i}", tb2out_1.extent[i]))
                config.append((f"strg_ub_tb_only_tb_read_addr_gen_1_strides_{i}", tb2out_1.out_data_stride[i]))
                config.append((f"strg_ub_tb_only_tb_read_sched_gen_1_sched_addr_gen_strides_{i}", tb2out_1.cyc_stride[i]))

        if "sram2tb_0" in root_node:
            sram2tb_0 = map_controller(extract_controller_json(root_node["sram2tb_0"]), "sram2tb_0")
            config.append(("strg_ub_sram_only_output_addr_gen_0_starting_addr", sram2tb_0.out_data_strt))
            config.append(("strg_ub_tb_only_tb_write_addr_gen_0_starting_addr", sram2tb_0.in_data_strt))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_0_enable", 1))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_0_sched_addr_gen_starting_addr", sram2tb_0.cyc_strt))
            config.append(("strg_ub_sram_tb_shared_loops_buf2out_autovec_read_0_dimensionality", sram2tb_0.dim))
            for i in range(sram2tb_0.dim):
                config.append((f"strg_ub_sram_tb_shared_loops_buf2out_autovec_read_0_ranges_{i}", sram2tb_0.extent[i]))
                config.append((f"strg_ub_sram_only_output_addr_gen_0_strides_{i}", sram2tb_0.out_data_stride[i]))
                config.append((f"strg_ub_sram_tb_shared_output_sched_gen_0_sched_addr_gen_strides_{i}", sram2tb_0.cyc_stride[i]))
                config.append((f"strg_ub_tb_only_tb_write_addr_gen_0_strides_{i}", sram2tb_0.in_data_stride[i]))

        if "sram2tb_1" in root_node:
            sram2tb_1 = map_controller(extract_controller_json(root_node["sram2tb_1"]), "sram2tb_1")
            config.append(("strg_ub_sram_only_output_addr_gen_1_starting_addr", sram2tb_1.out_data_strt))
            config.append(("strg_ub_tb_only_tb_write_addr_gen_1_starting_addr", sram2tb_1.in_data_strt))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_1_enable", 1))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_1_sched_addr_gen_starting_addr", sram2tb_1.cyc_strt))
            config.append(("strg_ub_sram_tb_shared_loops_buf2out_autovec_read_1_dimensionality", sram2tb_1.dim))
            for i in range(sram2tb_1.dim):
                config.append((f"strg_ub_sram_tb_shared_loops_buf2out_autovec_read_1_ranges_{i}", sram2tb_1.extent[i]))
                config.append((f"strg_ub_sram_only_output_addr_gen_1_strides_{i}", sram2tb_1.out_data_stride[i]))
                config.append((f"strg_ub_sram_tb_shared_output_sched_gen_1_sched_addr_gen_strides_{i}", sram2tb_1.cyc_stride[i]))
                config.append((f"strg_ub_tb_only_tb_write_addr_gen_1_strides_{i}", sram2tb_1.in_data_stride[i]))

        if "stencil_valid" in root_node:
            stencil_valid = map_controller(extract_controller_json(root_node["stencil_valid"]), "stencil_valid")
            # Check actual stencil valid property of hardware before programming
            if self.stencil_valid:
                config.append((f"stencil_valid_sched_gen_enable", 1))
                config.append((f"stencil_valid_sched_gen_sched_addr_gen_starting_addr", stencil_valid.cyc_strt))
                config.append((f"loops_stencil_valid_dimensionality", stencil_valid.dim))
                for i in range(stencil_valid.dim):
                    config.append((f"loops_stencil_valid_ranges_{i}", stencil_valid.extent[i]))
                    config.append((f"stencil_valid_sched_gen_sched_addr_gen_strides_{i}", stencil_valid.cyc_stride[i]))

        # Control Signals...
        # Set the mode and activate the tile...
        config.append(("flush_reg_sel", 0))  # 1
        config.append(("flush_reg_value", 0))  # 1
        config.append(("mode", 0))  # 2
        config.append(("tile_en", 1))  # 1

        # TODO: Maybe need to check if size 1?
        for i in range(input_ports):
            config.append((f"ren_in_{i}_reg_sel", 1))
            config.append((f"ren_in_{i}_reg_value", 0))

        for i in range(output_ports):
            config.append((f"wen_in_{i}_reg_sel", 1))
            config.append((f"wen_in_{i}_reg_value", 0))

        return trim_config_list(flattened, config)

    def get_static_bitstream(self,
                             config_path,
                             in_file_name,
                             out_file_name):

        input_ports = 1
        output_ports = 1

        config = []

        controllers = ["in2agg_0", "in2agg_1", "agg2sram_0", "agg2sram_1", "sram2tb_0", "sram2tb_1", "tb2out_0", "tb2out_1"]

        controller_objs = [None] * len(controllers)
        for i in range(len(controllers)):
            c = controllers[i]
            in_path = config_path + '/' + in_file_name + c + '.csv'
            out_path = config_path + '/' + out_file_name + c + '.csv'

            if os.path.isfile(in_path):
                controller_objs[i] = map_controller(extract_controller(in_path), c)
            elif os.path.isfile(out_path):
                controller_objs[i] = map_controller(extract_controller(out_path), c)
            else:
                print(f"No {c} file provided. Is this expected?")

        in2agg_0, in2agg_1, agg2sram_0, agg2sram_1, \
            sram2tb_0, sram2tb_1, tb2out_0, tb2out_1 = \
            controller_objs

        # Getting bitstreams is a little unwieldy due to fault (or its underlying implementation) not
        # handling arrays in the interface.
        # To alleviate this, we create the flattened wrapper so we can query widths of config
        # registers and trim values to their bitwidths...
        flattened = create_wrapper_flatten(self.internal_generator.clone(),
                                           self.name + "_W")

        config = []

        # control signals
        config.append(("flush_reg_sel", 0))
        config.append(("flush_reg_value", 0))

        # set the mode and activate the tile
        config.append(("mode", 0))
        config.append(("tile_en", 1))

        # TODO: Maybe need to check if size 1?
        for i in range(input_ports):
            config.append((f"ren_in_{i}_reg_sel", 1))
            config.append((f"ren_in_{i}_reg_value", 0))

        for i in range(output_ports):
            config.append((f"wen_in_{i}_reg_sel", 1))
            config.append((f"wen_in_{i}_reg_value", 0))

        # Check the hardware if it supports stencil valid
        if self.stencil_valid:
            cfg_path = config_path + '/' + 'stencil_valid.csv'
            # Check if the stencil valid file exists...if it doesn't we just won't program it
            if os.path.exists(cfg_path):
                stcl_valid = map_controller(extract_controller(cfg_path), "stencil_valid")
                config.append((f"loops_stencil_valid_dimensionality", stcl_valid.dim))
                config.append((f"stencil_valid_sched_gen_enable", 1))
                config.append((f"stencil_valid_sched_gen_sched_addr_gen_starting_addr", stcl_valid.cyc_strt))
                for i in range(stcl_valid.dim):
                    config.append((f"loops_stencil_valid_ranges_{i}", stcl_valid.extent[i]))
                    config.append((f"stencil_valid_sched_gen_sched_addr_gen_strides_{i}", stcl_valid.cyc_stride[i]))
            else:
                print("No configuration file provided for stencil valid...are you expecting one to exist?")
                print(f"Bogus stencil valid path: {cfg_path}")

        if in2agg_0 is not None:
            config.append(("strg_ub_agg_only_agg_write_addr_gen_0_starting_addr", in2agg_0.in_data_strt))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_0_enable", 1))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_0_sched_addr_gen_starting_addr", in2agg_0.cyc_strt))
            config.append(("strg_ub_agg_only_loops_in2buf_0_dimensionality", in2agg_0.dim))
            for i in range(in2agg_0.dim):
                config.append((f"strg_ub_agg_only_loops_in2buf_0_ranges_{i}", in2agg_0.extent[i]))
                config.append((f"strg_ub_agg_only_agg_write_addr_gen_0_strides_{i}", in2agg_0.in_data_stride[i]))
                config.append((f"strg_ub_agg_only_agg_write_sched_gen_0_sched_addr_gen_strides_{i}", in2agg_0.cyc_stride[i]))

        if in2agg_1 is not None:
            config.append(("strg_ub_agg_only_agg_write_addr_gen_1_starting_addr", in2agg_1.in_data_strt))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_1_enable", 1))
            config.append(("strg_ub_agg_only_agg_write_sched_gen_1_sched_addr_gen_starting_addr", in2agg_1.cyc_strt))
            config.append(("strg_ub_agg_only_loops_in2buf_1_dimensionality", in2agg_1.dim))
            for i in range(in2agg_1.dim):
                config.append((f"strg_ub_agg_only_loops_in2buf_1_ranges_{i}", in2agg_1.extent[i]))
                config.append((f"strg_ub_agg_only_agg_write_addr_gen_1_strides_{i}", in2agg_1.in_data_stride[i]))
                config.append((f"strg_ub_agg_only_agg_write_sched_gen_1_sched_addr_gen_strides_{i}", in2agg_1.cyc_stride[i]))

        if agg2sram_0 is not None:
            config.append(("strg_ub_agg_sram_shared_loops_in2buf_autovec_write_0_dimensionality", agg2sram_0.dim))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_0_enable", 1))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_0_sched_addr_gen_starting_addr", agg2sram_0.cyc_strt))
            config.append(("strg_ub_agg_only_agg_read_addr_gen_0_starting_addr", agg2sram_0.out_data_strt))
            config.append(("strg_ub_sram_only_input_addr_gen_0_starting_addr", agg2sram_0.in_data_strt))
            for i in range(agg2sram_0.dim):
                config.append((f"strg_ub_agg_only_agg_read_addr_gen_0_strides_{i}", agg2sram_0.out_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_loops_in2buf_autovec_write_0_ranges_{i}", agg2sram_0.extent[i]))
                config.append((f"strg_ub_sram_only_input_addr_gen_0_strides_{i}", agg2sram_0.in_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_agg_read_sched_gen_0_sched_addr_gen_strides_{i}", agg2sram_0.cyc_stride[i]))

        if agg2sram_1 is not None:
            config.append(("strg_ub_agg_sram_shared_loops_in2buf_autovec_write_1_dimensionality", agg2sram_1.dim))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_1_enable", 1))
            config.append(("strg_ub_agg_sram_shared_agg_read_sched_gen_1_sched_addr_gen_starting_addr", agg2sram_1.cyc_strt))
            config.append(("strg_ub_agg_only_agg_read_addr_gen_1_starting_addr", agg2sram_1.out_data_strt))
            config.append(("strg_ub_sram_only_input_addr_gen_1_starting_addr", agg2sram_1.in_data_strt))
            for i in range(agg2sram_1.dim):
                config.append((f"strg_ub_agg_only_agg_read_addr_gen_1_strides_{i}", agg2sram_1.out_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_loops_in2buf_autovec_write_1_ranges_{i}", agg2sram_1.extent[i]))
                config.append((f"strg_ub_sram_only_input_addr_gen_1_strides_{i}", agg2sram_1.in_data_stride[i]))
                config.append((f"strg_ub_agg_sram_shared_agg_read_sched_gen_1_sched_addr_gen_strides_{i}", agg2sram_1.cyc_stride[i]))

        if sram2tb_0 is not None:
            config.append(("strg_ub_sram_only_output_addr_gen_0_starting_addr", sram2tb_0.out_data_strt))
            config.append(("strg_ub_tb_only_tb_write_addr_gen_0_starting_addr", sram2tb_0.in_data_strt))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_0_enable", 1))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_0_sched_addr_gen_starting_addr", sram2tb_0.cyc_strt))
            config.append(("strg_ub_sram_tb_shared_loops_buf2out_autovec_read_0_dimensionality", sram2tb_0.dim))
            for i in range(sram2tb_0.dim):
                config.append((f"strg_ub_sram_tb_shared_loops_buf2out_autovec_read_0_ranges_{i}", sram2tb_0.extent[i]))
                config.append((f"strg_ub_sram_only_output_addr_gen_0_strides_{i}", sram2tb_0.out_data_stride[i]))
                config.append((f"strg_ub_sram_tb_shared_output_sched_gen_0_sched_addr_gen_strides_{i}", sram2tb_0.cyc_stride[i]))
                config.append((f"strg_ub_tb_only_tb_write_addr_gen_0_strides_{i}", sram2tb_0.in_data_stride[i]))

        if sram2tb_1 is not None:
            config.append(("strg_ub_sram_only_output_addr_gen_1_starting_addr", sram2tb_1.out_data_strt))
            config.append(("strg_ub_tb_only_tb_write_addr_gen_1_starting_addr", sram2tb_1.in_data_strt))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_1_enable", 1))
            config.append(("strg_ub_sram_tb_shared_output_sched_gen_1_sched_addr_gen_starting_addr", sram2tb_1.cyc_strt))
            config.append(("strg_ub_sram_tb_shared_loops_buf2out_autovec_read_1_dimensionality", sram2tb_1.dim))
            for i in range(sram2tb_1.dim):
                config.append((f"strg_ub_sram_tb_shared_loops_buf2out_autovec_read_1_ranges_{i}", sram2tb_1.extent[i]))
                config.append((f"strg_ub_sram_only_output_addr_gen_1_strides_{i}", sram2tb_1.out_data_stride[i]))
                config.append((f"strg_ub_sram_tb_shared_output_sched_gen_1_sched_addr_gen_strides_{i}", sram2tb_1.cyc_stride[i]))
                config.append((f"strg_ub_tb_only_tb_write_addr_gen_1_strides_{i}", sram2tb_1.in_data_stride[i]))

        if tb2out_0 is not None:
            config.append((f"strg_ub_tb_only_tb_read_addr_gen_0_starting_addr", tb2out_0.out_data_strt))
            config.append((f"strg_ub_tb_only_tb_read_sched_gen_0_enable", 1))
            config.append((f"strg_ub_tb_only_tb_read_sched_gen_0_sched_addr_gen_starting_addr", tb2out_0.cyc_strt))
            config.append((f"strg_ub_tb_only_loops_buf2out_read_0_dimensionality", tb2out_0.dim))
            for i in range(tb2out_0.dim):
                config.append((f"strg_ub_tb_only_loops_buf2out_read_0_ranges_{i}", tb2out_0.extent[i]))
                config.append((f"strg_ub_tb_only_tb_read_addr_gen_0_strides_{i}", tb2out_0.out_data_stride[i]))
                config.append((f"strg_ub_tb_only_tb_read_sched_gen_0_sched_addr_gen_strides_{i}", tb2out_0.cyc_stride[i]))

        if tb2out_1 is not None:
            config.append((f"strg_ub_tb_only_tb_read_addr_gen_1_starting_addr", tb2out_1.out_data_strt))
            config.append((f"strg_ub_tb_only_tb_read_sched_gen_1_enable", 1))
            config.append((f"strg_ub_tb_only_tb_read_sched_gen_1_sched_addr_gen_starting_addr", tb2out_1.cyc_strt))
            config.append((f"strg_ub_tb_only_loops_buf2out_read_1_dimensionality", tb2out_1.dim))
            for i in range(tb2out_1.dim):
                config.append((f"strg_ub_tb_only_loops_buf2out_read_1_ranges_{i}", tb2out_1.extent[i]))
                config.append((f"strg_ub_tb_only_tb_read_addr_gen_1_strides_{i}", tb2out_1.out_data_stride[i]))
                config.append((f"strg_ub_tb_only_tb_read_sched_gen_1_sched_addr_gen_strides_{i}", tb2out_1.cyc_stride[i]))

        return trim_config_list(flattened, config)


# formal module functions
def get_formal_module(module):
    lake_dut, need_config_lift, use_sram_stub, tsmc_info = \
        get_lake_dut(module,
                     # need to lift config regs after generator cuts
                     do_config_lift=False)

    # cuts for modular formal solving
    if module == "agg":
        cut_generator(lake_dut["strg_ub"]["sram_only"])
        cut_generator(lake_dut["strg_ub"]["sram_tb_shared"])
        cut_generator(lake_dut["strg_ub"]["tb_only"])
    elif module == "sram":
        cut_generator(lake_dut["strg_ub"]["agg_only"])
        cut_generator(lake_dut["strg_ub"]["tb_only"])
    elif module == "tb":
        cut_generator(lake_dut["strg_ub"]["agg_only"])
        cut_generator(lake_dut["strg_ub"]["agg_sram_shared"])
        cut_generator(lake_dut["strg_ub"]["sram_only"])
    else:
        print("Error! Invalid module name given...must be one of agg, sram, or tb. Cuts not performed.")
        return lake_dut, need_config_lift, use_sram_stub, tsmc_info

    # config regs pass (needs to be after generator cuts)
    lift_config_reg(lake_dut.internal_generator)
    need_config_lift = False

    # extract formal annotation after config regs have been lifted up
    extract_formal_annotation(lake_dut, f"{module}_lake_top_annotation.txt", module)

    return lake_dut, need_config_lift, use_sram_stub, tsmc_info


def get_lake_dut(formal_module=None,
                 in_ports=2,
                 out_ports=2,
                 # no stencil valid needed for formal problems
                 stencil_valid=False,
                 tsmc_info=SRAMMacroInfo("tsmc_name"),
                 use_sram_stub=True,
                 fifo_mode=True,
                 mem_width=64,
                 do_config_lift=True):

    lake_dut = LakeTop(mem_width=mem_width,
                       interconnect_input_ports=in_ports,
                       interconnect_output_ports=out_ports,
                       sram_macro_info=tsmc_info,
                       use_sram_stub=use_sram_stub,
                       fifo_mode=fifo_mode,
                       add_clk_enable=True,
                       add_flush=True,
                       stencil_valid=stencil_valid,
                       formal_module=formal_module,
                       do_config_lift=do_config_lift)

    print(f"Supports Stencil Valid: {lake_dut.supports('stencil_valid')}")

    # if do_config_lift, then do not need_config_lift later
    return lake_dut, not do_config_lift, use_sram_stub, tsmc_info


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='LakeTop')
    parser.add_argument("-f",
                        help="optional: will generate verilog, annotation file, and dim to strides/range mapping collateral to solve a formal problem. must provide module to solve for, default = agg",
                        default="agg")

    args = parser.parse_args()

    need_config_lift = True

    # normal generation
    if args.f is None:
        prefix = ""
        lake_dut, need_config_lift, use_sram_stub, tsmc_info = get_lake_dut()
    # optional: to add generator cuts for formal module verilog + annotations
    else:
        module = args.f
        lake_dut, need_config_lift, use_sram_stub, tsmc_info = get_formal_module(module)
        prefix = f"{module}_"

    # config lift happens in all possible cases by this point
    assert not need_config_lift

    sram_port_pass = change_sram_port_names(use_sram_stub=use_sram_stub, sram_macro_info=tsmc_info)
    # generate verilog
    verilog(lake_dut, filename=f"{prefix}lake_top.sv",
            optimize_if=False,
            additional_passes={"change sram port names": sram_port_pass})



from kratos import *
from functools import reduce
import operator
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint


class ForLoop(Generator):
    '''
    Generate addresses for a single port
    '''
    def __init__(self,
                 iterator_support=6,
                 config_width=16):

        super().__init__(f"for_loop_{iterator_support}_{config_width}", debug=True)

        self.iterator_support = iterator_support
        self.config_width = config_width
        # Create params for instancing this module...
        self.iterator_support_par = self.param("ITERATOR_SUPPORT", clog2(iterator_support) + 1, value=self.iterator_support)
        self.config_width_par = self.param("CONFIG_WIDTH", clog2(config_width) + 1, value=self.config_width)

        # PORT DEFS: begin

        # INPUTS
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._ranges = self.input("ranges", self.config_width,
                                  size=self.iterator_support,
                                  packed=True, explicit_array=True)
        self._ranges.add_attribute(ConfigRegAttr("Ranges of address generator"))
        self._ranges.add_attribute(FormalAttr(f"{self._ranges.name}", FormalSignalConstraint.SOLVE))

        self._dimensionality = self.input("dimensionality", 1 + clog2(self.iterator_support))
        self._dimensionality.add_attribute(ConfigRegAttr("Dimensionality of address generator"))
        self._dimensionality.add_attribute(FormalAttr(f"{self._dimensionality.name}", FormalSignalConstraint.SOLVE))

        self._step = self.input("step", 1)
        # OUTPUTS

        # PORT DEFS: end

        # LOCAL VARIABLES: begin
        self._dim_counter = self.var("dim_counter", self.config_width,
                                     size=self.iterator_support,
                                     packed=True,
                                     explicit_array=True)

        self._strt_addr = self.var("strt_addr", self.config_width)

        self._counter_update = self.var("counter_update", 1)
        self._calc_addr = self.var("calc_addr", self.config_width)

        self._max_value = self.var("max_value", self.iterator_support)
        self._mux_sel = self.var("mux_sel", max(clog2(self.iterator_support), 1))
        self._mux_sel_out = self.output("mux_sel_out", max(clog2(self.iterator_support), 1))
        self.wire(self._mux_sel_out, self._mux_sel)
        # LOCAL VARIABLES: end
        # GENERATION LOGIC: begin
        self._done = self.var("done", 1)
        self._clear = self.var("clear", self.iterator_support)
        self._inc = self.var("inc", self.iterator_support)

        self._inced_cnt = self.var("inced_cnt", self._dim_counter.width)
        self.wire(self._inced_cnt, self._dim_counter[self._mux_sel] + 1)
        # Next_max_value
        self._maxed_value = self.var("maxed_value", 1)
        self.wire(self._maxed_value, (self._dim_counter[self._mux_sel] ==
                                      self._ranges[self._mux_sel]) & self._inc[self._mux_sel])

        self.add_code(self.set_mux_sel)
        for i in range(self.iterator_support):
            self.add_code(self.set_clear, idx=i)
            self.add_code(self.set_inc, idx=i)
            self.add_code(self.dim_counter_update, idx=i)
            self.add_code(self.max_value_update, idx=i)
        # GENERATION LOGIC: end

        self._restart = self.output("restart", 1)
        self.wire(self._restart, self._step & (~self._done))

    @always_comb
    # Find lowest ready
    def set_mux_sel(self):
        self._mux_sel = 0
        self._done = 0
        for i in range(self.iterator_support):
            if ~self._done:
                if ~self._max_value[i] & (i < self._dimensionality):
                    self._mux_sel = i
                    self._done = 1

    @always_comb
    def set_clear(self, idx):
        self._clear[idx] = 0
        if ((idx < self._mux_sel) | (~self._done)) & self._step:
            self._clear[idx] = 1

    @always_comb
    def set_inc(self, idx):
        self._inc[idx] = 0
        # We always increment the innermost, and then have priority
        # on clear in the flops.
        if (const(idx, 5) == 0) & self._step & (idx < self._dimensionality):
            self._inc[idx] = 1
        elif (idx == self._mux_sel) & self._step & (idx < self._dimensionality):
            self._inc[idx] = 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def dim_counter_update(self, idx):
        if ~self._rst_n:
            self._dim_counter[idx] = 0
        else:
            if self._clear[idx]:
                self._dim_counter[idx] = 0
            elif self._inc[idx]:
                self._dim_counter[idx] = self._inced_cnt

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def max_value_update(self, idx):
        if ~self._rst_n:
            self._max_value[idx] = 0
        else:
            if self._clear[idx]:
                self._max_value[idx] = 0
            elif self._inc[idx]:
                self._max_value[idx] = self._maxed_value

    def get_iter(self):
        return self.iterator_support

    def get_cfg_width(self):
        return self.config_width


if __name__ == "__main__":
    it_support = 6
    cf_width = 16
    db_dut = ForLoop(iterator_support=it_support,
                     config_width=cf_width)
    verilog(db_dut, filename=f"for_loop_{it_support}.sv")



from kratos import *
from lake.top.lake_top import LakeTop
from lake.utils.sram_macro import SRAMMacroInfo
from lake.passes.passes import lift_config_reg

import kratos as kts


class LakeChain(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=2,
                 input_iterator_support=6,  # Addr Controllers
                 output_iterator_support=6,
                 interconnect_input_ports=1,  # Connection to int
                 interconnect_output_ports=3,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 use_sram_stub=1,
                 sram_macro_info=SRAMMacroInfo(),
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 max_agg_schedule=32,
                 input_max_port_sched=32,
                 output_max_port_sched=32,
                 align_input=1,
                 max_line_length=128,
                 max_tb_height=1,
                 tb_range_max=128,
                 tb_sched_max=64,
                 max_tb_stride=15,
                 num_tb=1,
                 tb_iterator_support=2,
                 multiwrite=1,
                 max_prefetch=64,
                 config_data_width=16,
                 config_addr_width=8,
                 num_tiles=2,
                 remove_tb=False,
                 fifo_mode=False,
                 add_clk_enable=False,
                 add_flush=False):
        super().__init__("LakeChain", debug=True)

        fw_int = int(mem_width / data_width)
        data_words_per_set = 2 ** config_addr_width
        sets = int((fw_int * mem_depth) / data_words_per_set)

        sets_per_macro = max(1, int(mem_depth / data_words_per_set))
        total_sets = max(1, banks * sets_per_macro)

        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._data_in = self.input("data_in",
                                   data_width,
                                   size=interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)
        self._addr_in = self.input("addr_in",
                                   data_width,
                                   size=interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        self._wen = self.input("wen", interconnect_input_ports)
        self._ren = self.input("ren", interconnect_output_ports)

        self._config_data_in = self.input("config_data_in",
                                          config_data_width)

        self._config_addr_in = self.input("config_addr_in",
                                          config_addr_width)

        self._config_data_out = self.output("config_data_out", config_data_width,
                                            size=(num_tiles, total_sets),
                                            explicit_array=True,
                                            packed=True)

        self._config_read = self.input("config_read", 1)
        self._config_write = self.input("config_write", 1)
        self._config_en = self.input("config_en", total_sets)

        self._data_out = self.output("data_out",
                                     data_width,
                                     size=(num_tiles, interconnect_output_ports),
                                     packed=True,
                                     explicit_array=True)

        self._data_out_inter = self.var("data_out_inter",
                                        data_width,
                                        size=(num_tiles, interconnect_output_ports),
                                        packed=True,
                                        explicit_array=True)

        self._valid_out = self.output("valid_out",
                                      interconnect_output_ports,
                                      size=num_tiles,
                                      packed=True,
                                      explicit_array=True)

        self._valid_out_inter = self.var("valid_out_inter",
                                         interconnect_output_ports,
                                         size=num_tiles,
                                         packed=True,
                                         explicit_array=True)

        self._enable_chain_output = self.input("enable_chain_output", 1)

        self._chain_data_out = self.output("chain_data_out",
                                           data_width,
                                           size=interconnect_output_ports,
                                           packed=True,
                                           explicit_array=True)

        self._chain_valid_out = self.output("chain_valid_out",
                                            interconnect_output_ports)

        self._tile_output_en = self.var("tile_output_en",
                                        1,
                                        size=(num_tiles, interconnect_output_ports),
                                        packed=True,
                                        explicit_array=True)

        self.is_valid_ = self.var("is_valid",
                                  1,
                                  size=interconnect_output_ports,
                                  packed=True,
                                  explicit_array=True)

        self.valids = self.var("valids",
                               clog2(num_tiles),
                               size=interconnect_output_ports,
                               packed=True,
                               explicit_array=True)

        for i in range(num_tiles):
            tile = LakeTop(data_width=data_width,
                           mem_width=mem_width,
                           mem_depth=mem_depth,
                           banks=banks,
                           input_iterator_support=input_iterator_support,
                           output_iterator_support=output_iterator_support,
                           interconnect_input_ports=interconnect_input_ports,
                           interconnect_output_ports=interconnect_output_ports,
                           mem_input_ports=mem_input_ports,
                           mem_output_ports=mem_output_ports,
                           use_sram_stub=use_sram_stub,
                           sram_macro_info=sram_macro_info,
                           read_delay=read_delay,
                           rw_same_cycle=rw_same_cycle,
                           agg_height=agg_height,
                           max_agg_schedule=max_agg_schedule,
                           input_max_port_sched=input_max_port_sched,
                           output_max_port_sched=output_max_port_sched,
                           align_input=align_input,
                           max_line_length=max_line_length,
                           max_tb_height=max_tb_height,
                           tb_range_max=tb_range_max,
                           tb_sched_max=tb_sched_max,
                           max_tb_stride=max_tb_stride,
                           num_tb=num_tb,
                           tb_iterator_support=tb_iterator_support,
                           multiwrite=multiwrite,
                           max_prefetch=max_prefetch,
                           config_data_width=config_data_width,
                           config_addr_width=config_addr_width,
                           num_tiles=num_tiles,
                           remove_tb=remove_tb,
                           fifo_mode=fifo_mode,
                           add_clk_enable=add_clk_enable,
                           add_flush=add_flush)

            self.add_child(f"tile_{i}", tile,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           enable_chain_output=self._enable_chain_output,
                           # tile index
                           chain_idx_input=i,
                           chain_idx_output=0,
                           tile_output_en=self._tile_output_en[i],
                           # broadcast input data to all tiles
                           data_in=self._data_in,
                           addr_in=self._addr_in,
                           wen=self._wen,
                           ren=self._ren,
                           config_data_in=self._config_data_in,
                           config_addr_in=self._config_addr_in,
                           config_data_out=self._config_data_out[i],
                           config_read=self._config_read,
                           config_write=self._config_write,
                           config_en=self._config_en,
                           # used if output chaining not enabled
                           data_out=self._data_out_inter[i],
                           valid_out=self._valid_out_inter[i],
                           # unused currently?
                           tile_en=1,
                           # UB mode
                           mode=0)

        self.add_code(self.set_data_out)
        self.add_code(self.set_valid_out)
        self.add_code(self.set_chain_outputs)

        # config regs
        lift_config_reg(self.internal_generator)

    @always_comb
    def set_data_out(self):
        self._data_out = self._data_out_inter

    @always_comb
    def set_chain_outputs(self):
        for i in range(interconnect_output_ports):
            self.valids[i] = 0
            self.is_valid_[i] = 0
        for i in range(num_tiles):
            for j in range(interconnect_output_ports):
                if (self._tile_output_en[i][j] == 1):
                    # keeps track of while tile has data/valid for this port
                    self.valids[j] = i
                    # indicates that this port value has been updated for some tile
                    self.is_valid_[j] = 1
        for i in range(interconnect_output_ports):
            if (self._enable_chain_output == 0) | (self.is_valid_[i] == 0):
                self._chain_data_out[i] = 0
                self._chain_valid_out[i] = 0
            else:
                self._chain_data_out[i] = self._data_out_inter[self.valids[i]][i]
                self._chain_valid_out[i] = self._valid_out_inter[self.valids[i]][i]

    @always_comb
    def set_valid_out(self):
        for i in range(num_tiles):
            for j in range(interconnect_output_ports):
                if self._tile_output_en[i][j] == 0:
                    self._valid_out[i][j] = 0
                elif self._enable_chain_output:
                    self._valid_out[i][j] = 0
                else:
                    self._valid_out[i][j] = self._valid_out_inter[i][j]


if __name__ == "__main__":
    dut = LakeChain(num_tiles=2)
    verilog(dut, filename="top_chain.sv",
            optimize_if=False)
